#!/usr/bin/env python
#
# Problem definition and original response:
# https://stackoverflow.com/a/17025929/2379433
# https://stackoverflow.com/a/32906273/2379433
#
# Author: Mike McKerns (mmckerns @caltech and @uqfoundation)
# Copyright (c) 2018-2022 The Uncertainty Quantification Foundation.
# License: 3-clause BSD.  The full license text is available at:
#  - https://github.com/uqfoundation/mystic/blob/master/LICENSE
"""
Maximize: f = 2*x[0]*x[1] + 2*x[0] - x[0]**2 - 2*x[1]**2

Subject to: -2*x[0] + 2*x[1] <= -2
             2*x[0] - 4*x[1] <= 0
               x[0]**3 -x[1] == 0

where: 0 <= x[0] <= inf
       1 <= x[1] <= inf
"""
import numpy as np
import mystic.symbolic as ms
import mystic.solvers as my
import mystic.math as mm

# generate constraints and penalty for a nonlinear system of equations 
ieqn = '''
   -2*x0 + 2*x1 <= -2
    2*x0 - 4*x1 <= 0'''
eqn = '''
     x0**3 - x1 == 0'''
cons = ms.generate_constraint(ms.generate_solvers(ms.simplify(eqn,target='x1')))
pens = ms.generate_penalty(ms.generate_conditions(ieqn), k=1e3)
bounds = [(0., None), (1., None)]

# get the objective
def objective(x, sign=1):
  x = np.asarray(x)
  return sign * (2*x[0]*x[1] + 2*x[0] - x[0]**2 - 2*x[1]**2)

# solve    
x0 = np.random.rand(2)
sol = my.fmin_powell(objective, x0, constraint=cons, penalty=pens, disp=True,
                     bounds=bounds, gtol=3, ftol=1e-6, full_output=True,
                     args=(-1,))

print('x* = %s; f(x*) = %s' % (sol[0], -sol[1]))



#!/usr/bin/env python
#
# Author: Mike McKerns (mmckerns @caltech and @uqfoundation)
# Copyright (c) 1997-2016 California Institute of Technology.
# Copyright (c) 2016-2022 The Uncertainty Quantification Foundation.
# License: 3-clause BSD.  The full license text is available at:
#  - https://github.com/uqfoundation/mystic/blob/master/LICENSE
"""
data structures for legacy data observations of lipschitz functions
"""
from numpy import inf, asarray
NULLSLOPE = inf

def _fails(filter, data=None):
  """filter a data array for failures using a boolean filter

  Inputs:
    filter -- a list of booleans (same size as data)
    data -- a list of data points

  Notes:
    if data=None, return the pairs of indices where data fails
  """
  from numpy import where, asarray
  failures = asarray(where(filter == False)).T
  # if data not given, then return indices for failure
  if not data: return failures
  # otherwise, return the list of 'bad' points
  failures = set( failures.flat )
  return asarray([data[i] for i in failures])

class lipschitzcone(list):
  """ Lipschitz double cone around a data point, with vertex and slope

 queries:
  vertex -- coordinates of lipschitz cone vertex
  slopes -- lipschitz slopes for the cone (should be same dimension as 'vertex')

 methods:
  contains -- return True if a given point is within the cone
  distance -- sum of lipschitz-weighted distance between a point and the vertex
"""

  def __init__(self, datapoint, slopes=None):
    self.vertex = datapoint
    if not slopes: slopes = [NULLSLOPE]*len(self.vertex.position)
    self.slopes = slopes
    return

  def __repr__(self):
    return "cone(%s @%s)" % (self.slopes, self.vertex.position)

  def distance(self, point):
    """ sum of lipschitz-weighted distance between a point and the vertex """
    x = point.position
    Z = self.vertex.position
    L = self.slopes
    return sum([L[i]*abs(x[i] - Z[i]) for i in range(len(x))])

  def contains(self, point):
    """ return True if a given point is within the cone """
    L = self.slopes
    if any([i == NULLSLOPE for i in L]): return True
    x = point.position
    Z = self.vertex.position
    if all([x[i] == Z[i] for i in range(len(x))]): return True
    y = point.value
    G = self.vertex.value
    return abs(y - G) <= self.distance(point)

  pass

class point(object):
  """ n-d data point with position and value but no id (i.e. 'raw')

 queries:
  p.value   --  returns value
  p.position  --  returns position
  p.rms  --  returns the square root of sum of squared position

 settings:
  p.value = v1  --  set the value
  p.position = (x1, x2, ..., xn)  --  set the position
"""

  def __init__(self, position, value):
    self.value = value
    self.position = position
    return
  
  def __repr__(self):
    return "(%s @%s)" % (self.value, self.position)

  def __rms(self): # square root of sum of squared positions
    from math import sqrt
    return sqrt(sum([i**2 for i in self.position]))

  ## compare
  def __eq__(self, other): # equal if value and position are equal
    if type(other) == type(self):
      return other.value == self.value and other.position == self.position
    return False

  def __ne__(self, other):
    return not other == self

  # sorts by rms(position) then by value... better by value first ???
  def __lt__(self, other):
    if type(other) == type(self):
      if self.position != other.position:
        return other.rms - self.rms > 0.0
      return other.value - self.value > 0.0
    return NotImplemented

  def __gt__(self, other):
    if type(other) == type(self):
      if self.position != other.position:
        return other.rms - self.rms < 0.0
      return other.value - self.value < 0.0
    return NotImplemented

  def __le__(self, other):
    if type(other) == type(self):
      if self.position != other.position:
        return other.rms - self.rms >= 0.0
      return other.value - self.value >= 0.0
    return NotImplemented

  def __ge__(self, other):
    if type(other) == type(self):
      if self.position != other.position:
        return other.rms - self.rms <= 0.0
      return other.value - self.value <= 0.0
    return NotImplemented
  ## end compare

  # interface
  rms = property(__rms)

  pass


class datapoint(object):
  """ n-d data point with position and value 

 queries:
  p.value   --  returns value
  p.position  --  returns position

 settings:
  p.value = v1  --  set the value
  p.position = (x1, x2, ..., xn)  --  set the position

 notes:
  a datapoint can have an assigned id and cone;
  also has utilities for comparison against other
  datapoints (intended for use in a dataset)
"""

  __hash_id = id  # save a copy of the builtin id function

  def __init__(self, position, value=None, id=None, lipschitz=None):
    self.raw = point([],None)
    if type(position) == type(self.raw):
      self.raw = position
    else:
      self.raw.position = position 
      if value is None: raise ValueError("value not set for the datapoint")
      self.raw.value = value
    if id is None: id = self.__hash_id(self)  #XXX: default id is hash id
    self.id = id
    self.cone = lipschitzcone(self,lipschitz)
    return

  # compare (defer to 'raw')
  def __eq__(self, other):
    if type(other) == type(self):
      return other.raw == self.raw and other.id == self.id  #XXX: consider id?
    return False

  def __ne__(self, other):
    return not other == self

  def __lt__(self, other):
    if type(other) == type(self):
      return self.raw < other.raw
    return NotImplemented

  def __gt__(self, other):
    if type(other) == type(self):
      return self.raw > other.raw
    return NotImplemented

  def __le__(self, other):
    if type(other) == type(self):
      if self.raw == other.raw:     #XXX: consider id?
        return self.id <= other.id  #XXX: consider id?
      return self.raw <= other.raw
    return NotImplemented

  def __ge__(self, other):
    if type(other) == type(self):
      if self.raw == other.raw:     #XXX: consider id?
        return self.id >= other.id  #XXX: consider id?
      return self.raw >= other.raw
    return NotImplemented
  # end compare

  # integrity
  def __is_duplicate(self, pt): # same raw & id
    return False if self is pt else self == pt

  def __is_repeat(self, pt): # same raw, different id
    return False if self == pt else self.raw == pt.raw
   #return self.id != pt.id and self == pt

  def __is_conflict(self, pt): # same id, different raw
    return False if self == pt else self.id == pt.id
   #return self != pt and self.id == pt.id

  def __is_collision(self, pt): # same position, different value
    return False if self.raw == pt.raw else self.position == pt.position

  def duplicates(self, pts):
    """ return True where a point exists with same 'raw' and 'id' """
    return [self.__is_duplicate(i) for i in pts]

  def repeats(self, pts):
    """ return True where a point exists with same 'raw' but different 'id' """
    return [self.__is_repeat(i) for i in pts]

  def conflicts(self, pts):
    """ return True where a point exists with same 'id' but different 'raw' """
    return [self.__is_conflict(i) for i in pts]

  def collisions(self, pts):
    """ return True where a point exists with same 'position' and different 'value' """
    return [self.__is_collision(i) for i in pts]
  # end integrity

  def __repr__(self):
    return "(%r: %r)" % (self.id, self.raw)

  def __position(self):
    return self.raw.position

  def __value(self):
    return self.raw.value

  def __set_position(self, position):
    self.raw.position = position
    return

  def __set_value(self, value):
    self.raw.value = value
    return

  # interface
  position = property(__position, __set_position)
  value = property(__value, __set_value)

  pass

class dataset(list):  #FIXME: should just accept datapoints
  """ a collection of data points
  s = dataset([point1, point2, ..., pointN])

 queries:
  s.values  --  returns list of values
  s.coords  --  returns list of positions
  s.ids  --  returns list of ids
  s.raw  --  returns list of points
  s.npts  --  returns the number of points
  s.lipschitz  --  returns list of lipschitz constants

 settings:
  s.lipschitz = [s1, s2, ..., sn]  --  sets lipschitz constants

 methods:
  short  -- check for shortness with respect to given data (or self)
  valid  -- check for validity with respect to given model
  update  -- update the positions and values in the dataset 
  load  -- load a list of positions and a list of values to the dataset 
  fetch  -- fetch the list of positions and the list of values in the dataset
  intersection  -- return the set intersection between self and query
  filter  -- return dataset entries where mask array is True 
  has_id  -- return True where dataset ids are in query
  has_position  -- return True where dataset coords are in query
  has_point  -- return True where dataset points are in query
  has_datapoint  -- return True where dataset entries are in query

 notes:
  - datapoints should not be edited; except possibly for id
  - assumes that s.n = len(s.coords) == len(s.values)
  - all datapoints in a dataset should have the same cone.slopes
"""
  # - when datapoint added to dataset, SHOULD build cone from dataset.lipschitz
  #   dataset.__set_id should check 'dataset.contains(id)', so no duplicate ids

  __name__ = None

  def short(self, data=None, L=None, blamelist=False, pairs=True, \
                                     all=False, raw=False, **kwds):
    """check for shortness with respect to given data (or self)

Args:
    data (list, default=None): a list of data points, or the dataset itself.
    L (float, default=None): the lipschitz constant, or the dataset's constant.
    blamelist (bool, default=False): if True, indicate the infeasible points.
    pairs (bool, default=True): if True, indicate indices of infeasible points.
    all (bool, default=False): if True, get results for each individual point.
    raw (bool, default=False): if False, get boolean results (i.e. non-float).
    tol (float, default=0.0): maximum acceptable deviation from shortness.
    cutoff (float, default=tol): zero out distances less than cutoff.

Notes:
    Each point x,y can be thought to have an associated double-cone with slope
    equal to the lipschitz constant. Shortness with respect to another point is
    defined by the first point not being inside the cone of the second. We can
    allow for some error in shortness, a short tolerance *tol*, for which the
    point x,y is some acceptable y-distance inside the cone. While very tightly
    related, *cutoff* and *tol* play distinct roles; *tol* is subtracted from
    calculation of the lipschitz_distance, while *cutoff* zeros out the value
    of any element less than the *cutoff*.
"""
    tol = kwds['tol'] if 'tol' in kwds else 0.0 # get tolerance in y
    # default is to zero out distances less than tolerance
    cutoff = kwds['cutoff'] if 'cutoff' in kwds else tol
    if cutoff is True: cutoff = tol
    elif cutoff is False: cutoff = None

    if L is None: L = self.lipschitz
    if data is None: data = self
    from mystic.math.distance import lipschitz_distance, is_feasible
    # calculate the shortness
    Rv = lipschitz_distance(L, self, data, **kwds)
    ld = is_feasible(Rv, cutoff)
    if raw:
      x = Rv
    else:
      x = ld
    if not blamelist: return ld.all() if not all else x
    if pairs: return _fails(ld)
    # else lookup failures
    return _fails(ld, data)

  def valid(self, model, blamelist=False, pairs=True, \
                         all=False, raw=False, **kwds):
    """check for validity with respect to given model

Args:
    model (func): the model function, ``y' = F(x')``.
    blamelist (bool, default=False): if True, indicate the infeasible points.
    pairs (bool, default=True): if True, indicate indices of infeasible points.
    all (bool, default=False): if True, get results for each individual point.
    raw (bool, default=False): if False, get boolean results (i.e. non-float).
    ytol (float, default=0.0): maximum acceptable difference ``|y - F(x')|``.
    xtol (float, default=0.0): maximum acceptable difference ``|x - x'|``.
    cutoff (float, default=ytol): zero out distances less than cutoff.
    hausdorff (bool, default=False): hausdorff ``norm``, where if given,
        then ``ytol = |y - F(x')| + |x - x'|/norm``.

Notes:
    *xtol* defines the n-dimensional base of a pilar of height *ytol*,
    centered at each point. The region inside the pilar defines the space
    where a "valid" model must intersect. If *xtol* is not specified, then
    the base of the pilar will be a dirac at ``x' = x``. This function
    performs an optimization for each ``x`` to find an appropriate ``x'``.

    *ytol* is a single value, while *xtol* is a single value or an iterable.
    *cutoff* takes a float or a boolean, where ``cutoff=True`` will set the
    value of *cutoff* to the default. Typically, the value of *cutoff* is
    *ytol*, 0.0, or None. *hausdorff* can be False (e.g. ``norm = 1.0``),
    True (e.g. ``norm = spread(x)``), or a list of points of ``len(x)``.

    While *cutoff* and *ytol* are very tightly related, they play a distinct
    role; *ytol* is used to set the optimization termination for an acceptable
    ``|y - F(x')|``, while *cutoff* is applied post-optimization.

    If we are using the *hausdorff* norm, then *ytol* will set the optimization
    termination for an acceptable ``|y - F(x')| + |x - x'|/norm``, where the
    ``x`` values are normalized by ``norm = hausdorff``.
"""
    ytol = kwds['ytol'] if 'ytol' in kwds else 0.0 # get tolerance in y
    # default is to zero out distances less than tolerance
    cutoff = kwds['cutoff'] if 'cutoff' in kwds else ytol
    if cutoff is True: cutoff = ytol
    elif cutoff is False: cutoff = None

    from mystic.math.distance import graphical_distance, is_feasible
    # calculate the model validity
    Rv = graphical_distance(model, self, **kwds)
    ld = is_feasible(Rv, cutoff)
    if raw:
      x = Rv
    else:
      x = ld
    if not blamelist: return ld.all() if not all else x
    if pairs: return _fails(ld)
    # else lookup failures
    return _fails(ld, self)

  def update(self, positions, values):#ids=None):# positions,values are iterable
    """ update the positions and values in the dataset 

Returns:
    self (dataset): the dataset itself

Notes:
    positions and values provided must be iterable
"""
    lip = self.lipschitz
    ids = self.ids +\
          [None] * max(0, min(len(positions), len(values)) - len(self.ids))
    z = list(zip(positions, values, ids))
    self[:len(z)] = [datapoint(i,j,id=k) for (i,j,k) in z]
    self.lipschitz = lip
    return self

  def load(self, positions, values, ids=[]): # positions,values are iterable
    """load a list of positions and a list of values to the dataset 

Returns:
    self (dataset): the dataset itself

Notes:
    positions and values provided must be iterable
"""
    z = list(zip(positions, values))
    self.extend([datapoint(i,j) for (i,j) in z])
    if ids: #XXX: must be at least as long as 'z'
      for i in range(len(z)):
        self[-len(z)+i].id = ids[i]
    return self

# def generate(self, model, positions, ids=[]):
#   """use a function to generate a dataset from the given list of positions"""
#   values = map(model, positions)    #XXX: readily extended to parallel
#   self.load(positions, values, ids)
#   return

# def evaluate(self, model):
#   """evaluate a function at dataset positions, overwriting existing values"""
#   values = map(model, self.coords) #XXX: readily extended to parallel
#   self.update(self.coords, values)
#   return

  def fetch(self):
    """fetch the list of positions and the list of values in the dataset"""
    return self.coords, self.values

  # integrity
  def __has_duplicate(self):
    return [any(i.duplicates(self)) for i in self]
   #return [i.duplicates(self) for i in self]

  def __has_repeat(self):
    return [any(i.repeats(self)) for i in self]

  def __has_conflict(self):
    return [any(i.conflicts(self)) for i in self]

  def __has_collision(self):
    return [any(i.collisions(self)) for i in self]

# def __has_duplicate(self, pt): #XXX XXX: checks if pt is duplicated in self
#   return True if self.count(pt) > 1 else False
#   # could return w/o duplicated by: dataset(set(self))

# def __has_conflict(self, pt): #XXX XXX: assumes repeated id not duplicate
#   return True if self.ids.count(pt.id) > 1 else False

# def __has_collision(self, pt): #XXX XXX: assumes repeated coord not duplicate
#   return True if self.coords.count(pt.position) > 1 else False

##def __collision(self, pt):
##  return not pt in self and pt.position in self.coords
  # end integrity

  def has_id(self, query): #FIXME: assume is iterable & appropriate
    """return True where dataset ids are in query

Notes:
    query must be iterable
"""
    return [i in query for i in self.ids]

  def has_position(self, query): #FIXME: assume is iterable & appropriate
    """return True where dataset coords are in query

Notes:
    query must be iterable
"""
    #XXX: allow query to contain lists, and not only tuples?
    return [i in query for i in self.coords]

  def has_point(self, query): #FIXME: assume is iterable & appropriate
    """return True where dataset points are in query

Notes:
    query must be iterable
"""
    return [i in query for i in self.raw]

  def has_datapoint(self, query): #FIXME: assume is iterable & appropriate
    """return True where dataset entries are in query

Notes:
    query must be iterable
"""
    return [i in query for i in self]

  def intersection(self, query):
    "return the set intersection between self and query"
    return dataset(set(self).intersection(query))

  def filter(self, mask): #XXX: assumes len(mask) = len(self); mask[i] is bool
    """return dataset entries where mask array is True 

Inputs:
    mask -- a boolean array of the same length as dataset
"""
    from numpy import array, where
    _self = array(self)
    return dataset(_self[where(mask)])

  def __data(self):
    return list(self)

  def __values(self):
    return [i.value for i in self]

  def __coords(self):
    return [i.position for i in self]

  def __ids(self):
    return [i.id for i in self]

  def __raw(self):
    return [i.raw for i in self]

  def __lipschitz(self):
    if not len(self): return []
#   # if not all datapoints have cones, throw error
#   lip = [self[i].cone == NULLCONE for i in range(len(self))]
#   if any(lip):
#     raise ValueError. "Lipschitz constants not set for the dataset."
    # if all cones don't have the same slopes, throw error
    lip = [self[i].cone.slopes != self[0].cone.slopes for i in range(len(self))]
    if any(lip):
      raise ValueError("Lipschitz constants not set for the dataset.")
    # return the lipschitz constants
    return self[0].cone.slopes

  def __npts(self):
    return len(self)

  def __repr__(self):
    return "dset(%r)" % ([pt for pt in self])

  def __set_lipschitz(self, slopes): #XXX: need len(slopes) == len(coords)
    for i in range(len(self)):
      self[i].cone = lipschitzcone(self[i], slopes)
    return

  def __set_ids(self, ids):
    for i in range(len(self)):
      self[i].id = ids[i]
    return

  def __set_values(self, values):
    for i in range(len(self)):
      self[i].value = values[i]
    return

  def __set_coords(self, coords):
    for i in range(len(self)):
      self[i].position = coords[i]
    return

  # interface
  values = property(__values)#, __set_values)
  coords = property(__coords)#, __set_coords)
  ids = property(__ids)#, __set_ids)
  raw = property(__raw)
  npts = property(__npts)
  lipschitz = property(__lipschitz, __set_lipschitz)
  duplicates = property(__has_duplicate)
  repeats = property(__has_repeat)
  conflicts = property(__has_conflict)
  collisions = property(__has_collision)
  pass


#######################################################
# legacy data file IO
#######################################################

def load_dataset(filename, filter=None):
  """ read dataset from selected file

  filename -- string name of dataset file
  filter -- tuple of points to select ('False' to ignore filter stored in file)
"""
  from os.path import split, splitext
  name = splitext(split(filename)[-1])[0]  # default name is filename
  lipschitz = None
  f = open(filename,"r")
  file = f.read()
  f.close()
  contents = file.split("\n")
  # parse file contents to get (i,id), cost, and parameters
  pt = []; _filter = None
  for line in contents[:-1]:
    if line.startswith("# data name"):
      name = eval(line.split(":")[-1])
    elif line.startswith("# lipschitz"):
      lipschitz = list( eval(line.split(":")[-1]) )
    elif line.startswith("# filter") and filter != False:
      _filter = eval(line.split(":")[-1])
    elif line.startswith("#"):
      pass
    else:
      data = line.split("   ")
      sid = eval(data[0])
      value = eval(data[1])
      coords = list(eval(data[-1]))
      pt.append(datapoint(list(coords), value, id=sid))

  # apply filter(s)
  from numpy import asarray
  if _filter is not None and filter != False:
    _filter = asarray(_filter)
    _filter = _filter[_filter < len(pt)]
    pt = [pt[i] for i in _filter]
  if filter is not None and filter != False:
    filter = asarray(filter)
    filter = filter[filter < len(pt)]
    pt = [pt[i] for i in filter]

  # build dataset
  mydataset = dataset(pt)
  mydataset.__name__ = name
  mydataset.lipschitz = lipschitz
  return mydataset

def save_dataset(data, filename='dataset.txt', filter=None, new=True):
  """ save dataset to selected file

  data -- data set
  filename -- string name of dataset file
  filter -- tuple, filter to apply to dataset upon reading
  new -- boolean, False if appending to existing file
"""
  import datetime
  if new: ind = 'w'
  else: ind = 'a'
  file = open(filename, ind)
  file.write("# %s\n" % datetime.datetime.now().ctime() )
  if data.__name__: file.write('# data name: "%s"\n' % data.__name__ )
  if data.lipschitz:
    file.write("# lipschitz: %s\n" % str(tuple(data.lipschitz)) )
  if filter is not None:
    file.write("# filter: %s\n" % str(filter))
  file.write("# ___id___  __value__  __coords__\n")
  for pt in data:
    x = "%s" % str(tuple(pt.position))
    y = "%s" % pt.value
    file.write("  '%s'     %s     %s\n" % (pt.id, y, x))
  file.close()
  return


if __name__ == '__main__':
  x = [1,1,0]; x3 = [1,1,1]; x4 = [0,0,0]
  y = 1; y2 = 2; y3 = 0; y4 = 2; y5 = 3
  print("creating a datapoint...")
  pt1 = datapoint(x,y,id='1')
  print(pt1)
  print("with: %s\n" % pt1.cone)

  print("creating a second datapoint...")
  pt2 = datapoint(x,y2,lipschitz=[1,1,1])
  print(pt2)
  print("with: %s\n" % pt2.cone)

  print("creating a third datapoint...")
  pt3 = datapoint(x3,y3,lipschitz=[1,1,1])
  print(pt3)
  print("with: %s\n" % pt3.cone)

  print("creating a fourth datapoint...")
  pt4 = datapoint(x4,y4,lipschitz=[0.25,0.25,0.25])
  print(pt4)
  print("with: %s\n" % pt4.cone)

  print("creating a fifth datapoint...")
  pt5 = datapoint(x4,y5,lipschitz=[0.25,0.25,0.25])
  print(pt5)
  print("with: %s\n" % pt5.cone)

  print("testing 'cone.contains'...")
  print("1st cone contains 1st, 2nd, 3rd, 4th, 5th point? %s, %s, %s, %s, %s" % (pt1.cone.contains(pt1), pt1.cone.contains(pt2), pt1.cone.contains(pt3), pt1.cone.contains(pt4), pt1.cone.contains(pt5)))
  print("2nd cone contains 1st, 2nd, 3rd, 4th, 5th point? %s, %s, %s, %s, %s" % (pt2.cone.contains(pt1), pt2.cone.contains(pt2), pt2.cone.contains(pt3), pt2.cone.contains(pt4), pt2.cone.contains(pt5)))
  print("3rd cone contains 1st, 2nd, 3rd, 4th, 5th point? %s, %s, %s, %s, %s" % (pt3.cone.contains(pt1), pt3.cone.contains(pt2), pt3.cone.contains(pt3), pt3.cone.contains(pt4), pt3.cone.contains(pt5)))
  print("4th cone contains 1st, 2nd, 3rd, 4th, 5th point? %s, %s, %s, %s, %s" % (pt4.cone.contains(pt1), pt4.cone.contains(pt2), pt4.cone.contains(pt3), pt4.cone.contains(pt4), pt4.cone.contains(pt5)))
  print("5th cone contains 1st, 2nd, 3rd, 4th, 5th point? %s, %s, %s, %s, %s\n" % (pt5.cone.contains(pt1), pt5.cone.contains(pt2), pt5.cone.contains(pt3), pt5.cone.contains(pt4), pt5.cone.contains(pt5)))

  print( "creating a dataset...")
  dset = dataset([pt1,pt2,pt3])
  print(dset)
  print("values: %s:" % dset.values)
  print("coords: %s:" % dset.coords)
  print("ids: %s" % dset.ids)
  print("raw: %s\n" % dset.raw)

  print("setting lipschitz constants for dataset...")
 #print(dset.lipschitz)
  dset.lipschitz = pt2.cone.slopes
  print("all lipschitz: %s" % dset.lipschitz)
  print("resulting cones: %s\n" % [p.cone for p in dset])


  ##### comparisons #####
  p0 = point([1,1,1],1)
  p1 = point([1,1,1],1)
  p2 = point([1,1,1],2)
  p3 = point([1,2,1],1)
  p4 = point([1,1,1],1)
  p5 = point([1,1,1],1)
  assert p0 == p1 and p0 != p2 and p0 != p3 and p0 != ([1,1,1],1)
  assert p2 > p1 and p3 > p1 and p2 < p3  #XXX: p3 > p2 ?

  d0 = datapoint(p0, id='0') # (original)
  d1 = datapoint(p1, id='1') #  conflict (?)
  d2 = datapoint(p2, id='2') #  collision
  d3 = datapoint(p3, id='3') #    ---
  d4 = datapoint(p4, id='0') #  duplicate
  d5 = datapoint(p5, id='3') #  conflict (?)
  assert d0 != d1 and d0 != d2 and d0 != d3 and d0 != p0
  assert d2 > d1 and d3 > d1 and d2 < d3  #XXX: d3 > d2 ?

  ##### intersections #####
  print("testing membership...")
  d = dataset([d0,d1,d2])
  print("point:")
  query = [p1,p2]
  print("%r in %r is %r" % (query, d.raw, d.filter(d.has_point(query)).raw))
  query = [p0,p3]
  print("%r in %r is %r" % (query, d.raw, d.filter(d.has_point(query)).raw))
  print("coords:")
  query = [p0.position, p2.position]
  print("%r in %r is %r" % (query, d.coords, d.filter(d.has_position(query)).coords))
  query = [p1.position, p3.position]
  print("%r in %r is %r" % (query, d.coords, d.filter(d.has_position(query)).coords))
  query = [p3.position]
  print("%r in %r is %r" % (query, d.coords, d.filter(d.has_position(query)).coords))
  print("id:")
  query = ['0','2']
  print("%r in %r is %r" % (query, d.ids, d.filter(d.has_id(query)).ids))
  query = ['1','3']
  print("%r in %r is %r" % (query, d.ids, d.filter(d.has_id(query)).ids))
  print("datapoint:")
  query = [d1,d2]
  print("%r in %r is %r" % (dataset(query).ids, d.ids, d.filter(d.has_datapoint(query)).ids))
  print("datapoint by intersection:")
  query = [d0,d2]
  print("%r in %r is %r" % (dataset(query).ids, d.ids, d.intersection(query).ids))

  ##### integrity #####
  print("testing integrity...")
  d = dataset([d0,d1,d2,d3,d4,d5])
  print("duplicates: %r" % d.duplicates)
  print("repeats: %r" % d.repeats)
  print("conflicts: %r" % d.conflicts)
  print("collisions: %r" % d.collisions)

  print("%r collides: %r" % (d[0], d[0].collisions(d)))
  print("%r collides: %r" % (d[1], d[1].collisions(d)))
  print("%r collides: %r" % (d[2], d[2].collisions(d)))
  print("%r collides: %r" % (d[3], d[3].collisions(d)))
  print("%r collides: %r" % (d[4], d[4].collisions(d)))
  print("%r collides: %r" % (d[5], d[5].collisions(d)))


# EOF



#!/usr/bin/env python
#
# Author: Patrick Hung (patrickh @caltech)
# Author: Mike McKerns (mmckerns @caltech and @uqfoundation)
# Copyright (c) 1997-2016 California Institute of Technology.
# Copyright (c) 2016-2022 The Uncertainty Quantification Foundation.
# License: 3-clause BSD.  The full license text is available at:
#  - https://github.com/uqfoundation/mystic/blob/master/LICENSE
"""
Support Vector Regression. Example 1
"""

from numpy import *
import matplotlib.pyplot as plt
from mystic.svr import *

# define the objective function to match standard QP solver
# (see: http://www.mathworks.com/help/optim/ug/quadprog.html)
def objective(x, Q, b):
    return 0.5 * dot(dot(x,Q),x) + dot(b,x)

# define the data points (linear data with uniform scatter)
x = arange(-5, 5.001); nx = x.size
y = x + 7*random.rand(nx)
N = 2*nx

# build the Kernel Matrix (with linear kernel)
# get the QP quadratic term
X = concatenate([x,-x])
lk = LinearKernel
Q = KernelMatrix(X, kernel=lk)
# get the QP linear term
Y = concatenate([y,-y])
svr_epsilon = 3
b = Y + svr_epsilon * ones(Y.size)

# build the constraints (y.T * x = 0.0)
# 1.0*x0 + 1.0*x1 + ... - 1.0*xN = 0.0
Aeq = concatenate([ones(nx), -ones(nx)]).reshape(1,N)
Beq = array([0.])
# set the bounds
lb = zeros(N)
ub = zeros(N) + 0.5

# build the constraints operator
from mystic.symbolic import linear_symbolic, solve, \
     generate_solvers as solvers, generate_constraint as constraint
constrain = linear_symbolic(Aeq,Beq)
constrain = constraint(solvers(solve(constrain,target=['x0'])))

from mystic import suppressed
@suppressed(1e-5)
def conserve(x):
    return constrain(x)

from mystic.monitors import VerboseMonitor
mon = VerboseMonitor(10)

# solve for alpha
from mystic.solvers import diffev
alpha = diffev(objective, list(zip(lb,.1*ub)), args=(Q,b), npop=N*3, gtol=200, \
               itermon=mon, \
               ftol=1e-5, bounds=list(zip(lb,ub)), constraints=conserve, disp=1)

print('solved x: %s' % alpha)
print("constraint A*x == 0: %s" % inner(Aeq, alpha))
print("minimum 0.5*x'Qx + b'*x: %s" % objective(alpha, Q, b))

# calculate support vectors and regression function
sv1 = SupportVectors(alpha[:nx])
sv2 = SupportVectors(alpha[nx:])
R = RegressionFunction(x, y, alpha, svr_epsilon, lk)

print('support vectors: %s %s' % (sv1, sv2))

# plot data
plt.plot(x, y, 'k+', markersize=10)

# plot regression function and support
plt.plot(x,R(x), 'k-')
plt.plot(x,R(x)-svr_epsilon, 'r--')
plt.plot(x,R(x)+svr_epsilon, 'g--')
plt.plot(x[sv1],y[sv1],'ro')
plt.plot(x[sv2],y[sv2],'go')
plt.show()

# end of file



#!/usr/bin/env python
#
# Problem definition and original response:
# https://stackoverflow.com/q/59813985/2379433
#
# Author: Mike McKerns (mmckerns @caltech and @uqfoundation)
# Copyright (c) 2020-2022 The Uncertainty Quantification Foundation.
# License: 3-clause BSD.  The full license text is available at:
#  - https://github.com/uqfoundation/mystic/blob/master/LICENSE
'''
solve:
     a system of inequalities (defined below) with unknowns xi,
     where i = range(1,11)

such that: 
     0 < x10 < x9 < x8 < x7 < x6 < x5 < x4 < x3 < x2 < x1,
     xi's are integers
'''

inequalities = '''
x1 > x2
x2 > x3
x3 > x4
x4 > x5
x5 > x6
x6 > x7
x7 > x8
x8 > x9
x9 > x10
x10 > 0
x5 < 3*x1/2 + x2 + x3/2 + x4/2 - x6/2
x5 < 3*x1/2 + 3*x2/2 + x3/2 + x8 - 3*x9/2 - x10/2
x4 > -x1 - 3*x2/2 + x5/2 - x8/2 + x10/2
x3 > -2*x1 - 3*x2/2 - x4 + x5 - x6/2 - x7/2 - x8/2 + x9 + 3*x10
x3 > -3*x1/2 - 3*x2/2 - x4/2 + 3*x5/2 - x6 + 3*x7/2 - x8/2 + x10
x3 > -3*x1/2 + x2 - x4/2 + 3*x5/2 - x6/2 + x7/2 - x8/2 + x10
x2 > -x1 - x3/2 - x4/2 - x6/4 + 3*x9/4 + x10/4
x2 > -x1 - x3/4 - x4/4 - x6/4 - x8/2 + 3*x9/4 + x10/2
x2 > -x1 + x4/4 + x5 - x6/4 - x7/4 - x8/4 + x9/4 + x10/4
x2 > -3*x1/4 - x3/2 - x4/2 - x6/2 - x7/4 - x8/2 + x9/2
x2 > -x1 - x3/2 - x4/4 - x6/4 - x8/4 + x9/4 + x10/4
x2 > -x1/2 + 3*x3/4 + x4/4 - x6/2 - x7/4 - x8/4 + x9/4 + x10/4
x1 > -x2 - x3/4 - x4/2 - x6/2 - x8/4
x1 > -x2 - x3/2 - x4/4 - x6/2 - x7/4 - x8/2 + x9/2
x1 > -x2 - x3/4 - x4/4 + x5/4 + 3*x7/4
x1 > -x2 - x3/2 - x4/2 - x6/2 - x7/4 - x8/4
x1 > -x2 - x3/2 - x4/4 - x6/4 - x7/4 - x8/4 + x9/4
'''

# test solutions
xA = [10,9,8,7,6,5,4,3,2,1]
xB = [1.1,2.3,3.7,4.3,5,6,7,8.932,9.0002,10]
xC = [64.251, 94., 62.123, 0.0, 41.234, 17.4, 0.0, 0.0, 81.341, 1.987]
x = xA, xB, xC


def failures(x):
    'count the number of failures in solvinge the inequalities'
    return tuple(eval(i, dict(zip(var,x))) for i in inequalities.strip().split('\n')).count(False)

def noints(x):
    'count the number of non-integer entries'
    return tuple(i == int(i) for i in x).count(False)


# solving the system of inequalities
import mystic as my
var = my.symbolic.get_variables(inequalities)
solve = my.symbolic.generate_constraint(my.symbolic.generate_solvers(inequalities, var), join=my.constraints.and_)

for xi in x:
    xo = xi.copy()
    y = solve(xo)
    assert not failures(y)


# building an integer constraint
ints = my.constraints.integers(float)(lambda x:x)

for xi in x:
    xo = xi.copy()
    y = ints(xo)
    assert not noints(y) #NOTE: yikes, poor english!


# now combine, with 'unique' and 'discrete' helping to force to integers
helper = my.constraints.and_(my.constraints.discrete(range(100))(lambda x:x), my.constraints.impose_unique(range(100))(lambda x:x))
intsolve = my.constraints.and_(ints, solve, helper, maxiter=1000)

for xi in x:
    xo = xi.copy()
    y = intsolve(xo)
    assert not failures(y) and not noints(y)




#!/usr/bin/env python
#
# Problem definition:
# A-R Hedar and M Fukushima, "Derivative-Free Filter Simulated Annealing
# Method for Constrained Continuous Global Optimization", Journal of
# Global Optimization, 35(4), 521-549 (2006).
# 
# Original Matlab code written by A. Hedar (Nov. 23, 2005)
# http://www-optima.amp.i.kyoto-u.ac.jp/member/student/hedar/Hedar_files/go.htm
# and ported to Python by Mike McKerns (December 2014)
#
# Author: Mike McKerns (mmckerns @caltech and @uqfoundation)
# Copyright (c) 1997-2016 California Institute of Technology.
# Copyright (c) 2016-2022 The Uncertainty Quantification Foundation.
# License: 3-clause BSD.  The full license text is available at:
#  - https://github.com/uqfoundation/mystic/blob/master/LICENSE

def objective(x):
    x0,x1,x2,x3,x4,x5,x6,x7,x8,x9 = x
    return x0**2 + x1**2 + x0*x1 - 14*x0 - 16*x1 + (x2-10)**2 + \
           4*(x3-5)**2 + (x4-3)**2 + 2*(x5-1)**2 + 5*x6**2 + \
           7*(x7-11)**2 + 2*(x8-10)**2 + (x9-7)**2 + 45.0

bounds = [(-10,10)]*10
# with penalty='penalty' applied, solution is:
xs = [2.171996, 2.363683, 8.773926, 5.095984, 0.9906548,
      1.430574, 1.321644, 9.828726, 8.280092, 8.375927]
ys = 24.3062091

from mystic.symbolic import generate_constraint, generate_solvers, simplify
from mystic.symbolic import generate_penalty, generate_conditions

equations = """
4.0*x0 + 5.0*x1 - 3.0*x6 + 9.0*x7 - 105.0 <= 0.0
10.0*x0 - 8.0*x1 - 17.0*x6 + 2.0*x7 <= 0.0
-8.0*x0 + 2.0*x1 + 5.0*x8 - 2.0*x9 - 12.0 <= 0.0
3.0*(x0-2)**2 + 4.0*(x1-3)**2 + 2.0*x2**2 - 7.0*x3 - 120.0 <= 0.0
5.0*x0**2 + 8.0*x1 + (x2-6)**2 - 2.0*x3 - 40.0 <= 0.0
0.5*(x0-8)**2 + 2.0*(x1-4)**2 + 3.0*x4**2 - x5 - 30.0 <= 0.0
x0**2 + 2.0*(x1-2)**2 - 2.0*x0*x1 + 14.0*x4 - 6.0*x5 <= 0.0
-3.0*x0 + 6.0*x1 + 12.0*(x8-8)**2 - 7.0*x9 <= 0.0
"""
cf = generate_constraint(generate_solvers(simplify(equations, target=['x5','x3'])))
pf = generate_penalty(generate_conditions(equations))



if __name__ == '__main__':
    x = [0]*len(xs)

    from mystic.solvers import fmin_powell
    from mystic.math import almostEqual
    from mystic.monitors import VerboseMonitor
    mon = VerboseMonitor(10)

    result = fmin_powell(objective, x0=x, bounds=bounds, penalty=pf, maxiter=1000, maxfun=100000, ftol=1e-12, xtol=1e-12, gtol=10, disp=False, full_output=True)

    assert almostEqual(result[0], xs, tol=1e-2)
    assert almostEqual(result[1], ys, rel=1e-2)



# EOF



#!/usr/bin/env python
#
# Author: Mike McKerns (mmckerns @caltech and @uqfoundation)
# Copyright (c) 2020-2022 The Uncertainty Quantification Foundation.
# License: 3-clause BSD.  The full license text is available at:
#  - https://github.com/uqfoundation/mystic/blob/master/LICENSE
"""
  Example applying mystic to scipy.optimize

  Minimize:
            f(x) = 3*A + 1e-6*A**3 + 2*B + 2e-6/(3*B**3) + C**2
            x = A,B,C

  Where: 
            A*B + C >= 1
            B <= A
            5 >= A >= -5
            5 >= B >= -5
            5 >= C >= -5
"""
import mystic as my
import mystic.symbolic as ms
import scipy.optimize as so

equations = """
A*B + C >= 1
B <= A
A >= -5
B >= -5
C >= -5
5 >= A
5 >= B
5 >= C
"""
var = list('ABC')
eqns = ms.simplify(equations, variables=var, all=True)
constrain = ms.generate_constraint(ms.generate_solvers(eqns, var), join=my.constraints.and_)

def objective(x):
    return 3*x[0] + 1.e-6*x[0]**3 + 2*x[1] + 2.e-6/3*x[1]**3 + x[2]**2

mon = my.monitors.Monitor()

def cost(x):
    kx = constrain(x)
    y = objective(kx)
    mon(kx,y)
    return y


if __name__ == '__main__':
    from mystic.math import almostEqual

    result = so.fmin(cost, [1,1,1], xtol=1e-6, ftol=1e-6, full_output=True, disp=False)
    # check results are consistent with monitor
    assert almostEqual(result[1], min(mon.y), rel=1e-2)

    # check results satisfy constraints
    A,B,C = result[0]
    print(dict(A=A, B=B, C=C))

    eps = 0.2 #XXX: give it some small wiggle room for small violations
    assert A*B + C >= 1-eps
    assert B <= A+eps
    assert (5+eps) >= A >= -(5+eps)
    assert (5+eps) >= B >= -(5+eps)
    assert (5+eps) >= C >= -(5+eps)



#!/usr/bin/env python
# ******NOTICE***************
# optimize.py module by Travis E. Oliphant
#
# You may copy and use this module as you see fit with no
# guarantee implied provided you keep this notice in all copies.
# *****END NOTICE************
#
# Forked by: Mike McKerns (February 2009)
# Author: Mike McKerns (mmckerns @caltech and @uqfoundation)
# Copyright (c) 2009-2016 California Institute of Technology.
# Copyright (c) 2016-2022 The Uncertainty Quantification Foundation.
# License: 3-clause BSD.  The full license text is available at:
#  - https://github.com/uqfoundation/mystic/blob/master/LICENSE
"""modified algorithms from local copy of scipy.optimize"""

#__all__ = ['fmin', 'fmin_powell', 'fmin_ncg', 'fmin_cg', 'fmin_bfgs',
#           'fminbound','brent', 'golden','bracket','rosen','rosen_der',
#           'rosen_hess', 'rosen_hess_prod', 'brute', 'approx_fprime',
#           'line_search', 'check_grad']

from mystic._scipy060optimize import *
from mystic._scipy060optimize import _cubicmin, _quadmin, _epsilon
from mystic._scipy060optimize import _linesearch_powell, _endprint

def line_search(f, myfprime, xk, pk, gfk, old_fval, old_old_fval,
                args=(), c1=1e-4, c2=0.9, amax=50, self=None):
    """Find alpha that satisfies strong Wolfe conditions.

    :Parameters:

        f : callable f(x,*args)
            Objective function.
        myfprime : callable f'(x,*args)
            Objective function gradient (can be None).
        xk : ndarray
            Starting point.
        pk : ndarray
            Search direction.
        gfk : ndarray
            Gradient value for x=xk (xk being the current parameter
            estimate).
        args : tuple
            Additional arguments passed to objective function.
        c1 : float
            Parameter for Armijo condition rule.
        c2 : float
            Parameter for curvature condition rule.

    :Returns:

        alpha0 : float
            Alpha for which ``x_new = x0 + alpha * pk``.
        fc : int
            Number of function evaluations made.
        gc : int
            Number of gradient evaluations made.

    :Notes:

        Uses the line search algorithm to enforce strong Wolfe
        conditions.  See Wright and Nocedal, 'Numerical Optimization',
        1999, pg. 59-60.

        For the zoom phase it uses an algorithm by [...].

    """

    global _ls_fc, _ls_gc, _ls_ingfk
    _ls_fc = 0
    _ls_gc = 0
    _ls_ingfk = None
    def phi(alpha):
        global _ls_fc
        _ls_fc += 1
        return f(xk+alpha*pk,*args)

    if isinstance(myfprime,type(())):
        def phiprime(alpha):
            global _ls_fc, _ls_ingfk
            _ls_fc += len(xk)+1
            eps = myfprime[1]
            fprime = myfprime[0]
            newargs = (f,eps) + args
            _ls_ingfk = fprime(xk+alpha*pk,*newargs)  # store for later use
            return numpy.dot(_ls_ingfk,pk)
    else:
        fprime = myfprime
        def phiprime(alpha):
            global _ls_gc, _ls_ingfk
            _ls_gc += 1
            _ls_ingfk = fprime(xk+alpha*pk,*args)  # store for later use
            return numpy.dot(_ls_ingfk,pk)

    alpha0 = 0
    phi0 = old_fval
    derphi0 = numpy.dot(gfk,pk)

    alpha1 = min(1.0,1.01*2*(phi0-old_old_fval)/derphi0)

    if alpha1 == 0:
        # This shouldn't happen. Perhaps the increment has slipped below
        # machine precision?  For now, set the return variables skip the
        # useless while loop, and raise warnflag=2 due to possible imprecision.
        alpha_star = None
        fval_star = old_fval
        old_fval = old_old_fval
        fprime_star = None

    phi_a1 = phi(alpha1)
    #derphi_a1 = phiprime(alpha1)  evaluated below

    phi_a0 = phi0
    derphi_a0 = derphi0

    i = 1
    maxiter = 10
    while 1:         # bracketing phase
        if alpha1 == 0:
            break
        if (phi_a1 > phi0 + c1*alpha1*derphi0) or \
           ((phi_a1 >= phi_a0) and (i > 1)):
            alpha_star, fval_star, fprime_star = \
                        zoom(alpha0, alpha1, phi_a0,
                             phi_a1, derphi_a0, phi, phiprime,
                             phi0, derphi0, c1, c2)
            break

        derphi_a1 = phiprime(alpha1)
        if (abs(derphi_a1) <= -c2*derphi0):
            alpha_star = alpha1
            fval_star = phi_a1
            fprime_star = derphi_a1
            break

        if (derphi_a1 >= 0):
            alpha_star, fval_star, fprime_star = \
                        zoom(alpha1, alpha0, phi_a1,
                             phi_a0, derphi_a1, phi, phiprime,
                             phi0, derphi0, c1, c2)
            break

        alpha2 = 2 * alpha1   # increase by factor of two on each iteration
        i = i + 1
        alpha0 = alpha1
        alpha1 = alpha2
        phi_a0 = phi_a1
        phi_a1 = phi(alpha1)
        derphi_a0 = derphi_a1

        # stopping test if lower function not found
        if (i > maxiter):
            alpha_star = alpha1
            fval_star = phi_a1
            fprime_star = None
            break

        if self._EARLYEXIT: # In case it gets stuck in this loop, 'exit' will work
            alpha_star = alpha1
            fval_star = phi_a1
            fprime_star = None
            break

    if fprime_star is not None:
        # fprime_star is a number (derphi) -- so use the most recently
        # calculated gradient used in computing it derphi = gfk*pk
        # this is the gradient at the next step no need to compute it
        # again in the outer loop.
        fprime_star = _ls_ingfk

    return alpha_star, _ls_fc, _ls_gc, fval_star, old_fval, fprime_star

def line_search_BFGS(f, xk, pk, gfk, old_fval, args=(), c1=1e-4, alpha0=1,\
                     self=None):
    """Minimize over alpha, the function ``f(xk+alpha pk)``.

    Uses the interpolation algorithm (Armiijo backtracking) as suggested by
    Wright and Nocedal in 'Numerical Optimization', 1999, pg. 56-57

    :Returns: (alpha, fc, gc)

    """

    xk = atleast_1d(xk)
    fc = 0
    phi0 = old_fval # compute f(xk) -- done in past loop
    phi_a0 = f(*((xk+alpha0*pk,)+args))
    fc = fc + 1
    derphi0 = numpy.dot(gfk,pk)

    if (phi_a0 <= phi0 + c1*alpha0*derphi0):
        return alpha0, fc, 0, phi_a0

    # Otherwise compute the minimizer of a quadratic interpolant:

    alpha1 = -(derphi0) * alpha0**2 / 2.0 / (phi_a0 - phi0 - derphi0 * alpha0)
    phi_a1 = f(*((xk+alpha1*pk,)+args))
    fc = fc + 1

    if (phi_a1 <= phi0 + c1*alpha1*derphi0): 
        return alpha1, fc, 0, phi_a1

    # Otherwise loop with cubic interpolation until we find an alpha which
    # satifies the first Wolfe condition (since we are backtracking, we will
    # assume that the value of alpha is not too small and satisfies the second
    # condition.

    while 1:       # we are assuming pk is a descent direction
        factor = alpha0**2 * alpha1**2 * (alpha1-alpha0)
        a = alpha0**2 * (phi_a1 - phi0 - derphi0*alpha1) - \
            alpha1**2 * (phi_a0 - phi0 - derphi0*alpha0)
        a = a / factor
        b = -alpha0**3 * (phi_a1 - phi0 - derphi0*alpha1) + \
            alpha1**3 * (phi_a0 - phi0 - derphi0*alpha0)
        b = b / factor

        alpha2 = (-b + numpy.sqrt(abs(b**2 - 3 * a * derphi0))) / (3.0*a)
        phi_a2 = f(*((xk+alpha2*pk,)+args))
        fc = fc + 1

        if (phi_a2 <= phi0 + c1*alpha2*derphi0):
            return alpha2, fc, 0, phi_a2

        if (alpha1 - alpha2) > alpha1 / 2.0 or (1 - alpha2/alpha1) < 0.96:
            alpha2 = alpha1 / 2.0

        alpha0 = alpha1
        alpha1 = alpha2
        phi_a0 = phi_a1
        phi_a1 = phi_a2

        # In case it gets stuck in this loop, to make 'exit' work.
        if self._EARLYEXIT:
            return alpha2, fc, 0, phi_a2

def approx_fprime(xk,f,epsilon,*args,**kwds):
    """Finite-difference approximation of the gradient of a scalar function.

    Parameters
    ----------
    xk : array_like
        The coordinate vector at which to determine the gradient of `f`.
    f : callable
        The function of which to determine the gradient (partial derivatives).
        Should take `xk` as first argument, other arguments to `f` can be
        supplied in ``*args``.  Should return a scalar, the value of the
        function at `xk`.
    epsilon : array_like
        Increment to `xk` to use for determining the function gradient.
        If a scalar, uses the same finite difference delta for all partial
        derivatives.  If an array, should contain one value per element of
        `xk`.
    f0 : float, optional
        Initial function value.
    \\*args : args, optional
        Any other arguments that are to be passed to `f`.

    Returns
    -------
    grad : ndarray
        The partial derivatives of `f` to `xk`.

    Notes
    -----
    The function gradient is determined by the forward finite difference
    formula::

                 f(xk[i] + epsilon[i]) - f(xk[i])
        f'[i] = ---------------------------------
                            epsilon[i]

    The main use of `approx_fprime` is in scalar function optimizers like
    `fmin_bfgs`, to determine numerically the Jacobian of a function.

    Examples
    --------
    >>> def func(x, c0, c1):
    ...     "Coordinate vector `x` should be an array of size two."
    ...     return c0 * x[0]**2 + c1*x[1]**2

    >>> x = np.ones(2)
    >>> c0, c1 = (1, 200)
    >>> eps = np.sqrt(np.finfo(float).eps)
    >>> approx_fprime(x, func, [eps, np.sqrt(200) * eps], c0, c1)
    array([   2.        ,  400.00004198])
    """
    xk = xk.copy()  #XXX: this line has been added for scipy_ncg
    f0 = kwds['f0'] if 'f0' in kwds else f(*((xk,)+args))
    grad = numpy.zeros((len(xk),), float)
    ei = numpy.zeros((len(xk),), float)
    for k in range(len(xk)):
        ei[k] = 1.0
        d = epsilon * ei
        grad[k] = (f(*((xk+d,)+args)) - f0) / d[k]
        ei[k] = 0.0
    return grad


if __name__ == "__main__":
    pass

# End of file



#!/usr/bin/env python
#
# Author: Mike McKerns (mmckerns @caltech and @uqfoundation)
# Copyright (c) 1997-2016 California Institute of Technology.
# Copyright (c) 2016-2022 The Uncertainty Quantification Foundation.
# License: 3-clause BSD.  The full license text is available at:
#  - https://github.com/uqfoundation/mystic/blob/master/LICENSE
"""
See test_rosenbrock.py.

This one uses Scipy's CG (Polak-Ribiere) plus viz via matplotlib

cg works well on this problem.
"""

import matplotlib.pyplot as plt
from test_rosenbrock import *
from numpy import log
from mystic._scipyoptimize import fmin_cg
import numpy
from mystic.tools import getch

def show():
    import matplotlib.pyplot as plt, Image
    plt.savefig('cg_rosenbrock_out',dpi=72)
    im = Image.open('cg_rosenbrock_out.png')
    im.show()
    return


def draw_contour():
    import numpy

    x, y = numpy.mgrid[-1:2.1:0.02,-0.1:2.1:0.02]
    c = 0*x
    s,t = x.shape
    for i in range(s):
        for j in range(t):
            xx,yy = x[i,j], y[i,j]
            c[i,j] = rosen([xx,yy])

    plt.contourf(x,y,log(c*20+1)+2,60)


def run_once(x0,x1,color='w'):
    sol = fmin_cg(rosen, [x0, x1], retall = True, full_output=1)
    xy = numpy.asarray(sol[-1])
    plt.plot(xy[:,0],xy[:,1],color+'-',linewidth=2)
    plt.plot(xy[:,0],xy[:,1],color+'o',markersize=6)
    return sol

    
if __name__ == '__main__':
    draw_contour()
    run_once(0.3,0.3,'k')
    run_once(0.5,1.3,'y')
    run_once(1.8,0.2,'w')
    try:
        show()
    except ImportError:
        plt.show()


# end of file



#!/usr/bin/env python
# 
# Author: Mike McKerns (mmckerns @uqfoundation)
# Copyright (c) 2020-2022 The Uncertainty Quantification Foundation.
# License: 3-clause BSD.  The full license text is available at:
#  - https://github.com/uqfoundation/mystic/blob/master/LICENSE
"""
calculate upper bound on expected value

Test function is y = F(x), where:
  F is defined in surrogate.py

model = F(x)

Calculate upper bound on E|model(x)|, where:
  x in [(20.0,150.0), (0.0,30.0), (2.1,2.8)]
  wx in [(0,1), (1,1), (1,1)]
  npts = [2, 1, 1] (i.e. two Dirac masses on x[0], one elsewhere)
  sum(wx[i]_j) for j in [0,npts], for each i
  E|model(x)| = 6.5 +/- 1.0

Solves for two scenarios of x that produce upper bound on E|model(x)|,
given the bounds, normalization, and moment constraints.
"""


if __name__ == '__main__':

    from spec3D import *
    from ouq import ExpectedValue
    from mystic.bounds import MeasureBounds
    from ouq_models import WrapModel
    from surrogate import marc_surr as toy; nx = 3; ny = None
    Ns = 25

    # build a model representing 'truth'
    nargs = dict(nx=nx, ny=ny, rnd=False)
    model = WrapModel('model', toy, **nargs) 

    # calculate upper bound on expected value, where F(x) has uncertainty
    bnd = MeasureBounds(xlb, xub, n=npts, wlb=wlb, wub=wub)
    b = ExpectedValue(model, bnd, constraint=scons, cvalid=is_cons, samples=Ns)
    b.upper_bound(axis=None, **param)
    print("upper bound per axis:")
    for axis,solver in b._upper.items():
        print("%s: %s @ %s" % (axis, -solver.bestEnergy, solver.bestSolution))



#!/usr/bin/env python
#
# Author: Mike McKerns (mmckerns @caltech and @uqfoundation)
# Author: Patrick Hung (patrickh @caltech)
# Copyright (c) 1997-2016 California Institute of Technology.
# Copyright (c) 2016-2022 The Uncertainty Quantification Foundation.
# License: 3-clause BSD.  The full license text is available at:
#  - https://github.com/uqfoundation/mystic/blob/master/LICENSE
__doc__ = _doc = """
Multi-minima example functions with vector outputs, which require
a 'reducing' function to provide scalar return values.

References:
    None
"""
from .abstract_model import AbstractFunction

from numpy import absolute as abs
from numpy import asarray
from numpy import sin, pi

class Wavy1(AbstractFunction): #XXX: not a standard test function...?
    __doc__ = \
    r"""a wavy1 function generator

A wavy1 function has a vector return value, and oscillates
similarly to x+\sin(x) in each direction. When a reduction
function, like 'numpy.add' is applied, the surface can be
visualized. The global minimum is at the center of a
cross-hairs running along x_i = -pi, with periodic local
minima in each direction.

The generated function f(x) requires len(x) > 0, and a
reducing function for use in most optimizers.
    """ + _doc
    def __init__(self, ndim=2): # is n-dimensional
        AbstractFunction.__init__(self, ndim=ndim)
        return

    def function(self,coeffs):
        r"""evaluates the wavy1 function for a list of coeffs

f(x) = \abs(x + 3*\sin(x + \pi) + \pi)

Inspect with mystic_model_plotter using::
    mystic.models.wavy1 -b "-20:20:.5, -20:20:.5" -d -r numpy.add

The minimum is f(x)=0.0 at x_i=-pi for all i"""
        x = asarray(coeffs) #XXX: must be numpy.array
        return abs(x+3.*sin(x+pi)+pi)

    minimizers = [-pi]
    pass


class Wavy2(AbstractFunction): #XXX: not a standard test function...?
    r"""a wavy2 function generator

A wavy2 function has a vector return value, and oscillates
similarly to \sin(x) in each direction. When a reduction
function, like 'numpy.add' is applied, the surface can be
visualized. There are degenerate global minima which are
periodic at 2*\pi, and similarly periodic local minima
at nearly the same location and magnitude.

The generated function f(x) requires len(x) > 0, and a
reducing function for use in most optimizers.
    """ + _doc
    def __init__(self, ndim=2): # is n-dimensional
        AbstractFunction.__init__(self, ndim=ndim)
        return

    def function(self,coeffs):
        r"""evaluates the wavy2 function for a list of coeffs

f(x) = 4*\sin(x)+\sin(4*x)+\sin(8*x)+\sin(16*x)+\sin(32*x)+\sin(64*x)

Inspect with mystic_model_plotter using::
    mystic.models.wavy2 -b "-10:10:.2, -10:10:.2" -d -r numpy.add

The function has degenerate global minima of f(x)=-6.987594
at x_i = 4.489843526 + 2*k*pi for all i, and k is an integer"""
        x = asarray(coeffs) #XXX: must be a numpy.array
        return 4*sin(x)+sin(4*x)+sin(8*x)+sin(16*x)+sin(32*x)+sin(64*x)

    minimizers = None #FIXME: degenerate minimum...
                      # minimum is ???
    pass

# cleanup
del _doc

# prepared instances
wavy1 = Wavy1().function
wavy2 = Wavy2().function

# End of file



#!/usr/bin/env python
#
# Author: Patrick Hung (patrickh @caltech)
# Author: Mike McKerns (mmckerns @caltech and @uqfoundation)
# Copyright (c) 1997-2016 California Institute of Technology.
# Copyright (c) 2016-2022 The Uncertainty Quantification Foundation.
# License: 3-clause BSD.  The full license text is available at:
#  - https://github.com/uqfoundation/mystic/blob/master/LICENSE
"""
Two mogi sources. Similar to test_mogi.py. (See that one first)

Reference:

[1] Mogi, K. Relations between the eruptions of various
volcanoes and the deformations of the ground surfaces around them, 
Bull. Earthquake. Res. Inst., 36, 99-134, 1958.
"""

from test_mogi import *

# Let the "actual parameters" be :
params0 = [1000.,-100., 10., .2]
params1 = [1500.,-400., 40., 1.5]

forward0 = ForwardMogiFactory(params0)
forward1 = ForwardMogiFactory(params1)

# The data to be "fitted" 
xstations = array([random.uniform(-500,500) for i in range(300)])+1250.
ystations =  0*xstations - 200.
stations  = array((xstations, ystations))

data = forward0(stations) + forward1(stations)
# noisy data, gaussian distribution
noise =  array([[random.normal(0,0.05e-5) for i in range(data.shape[1])] for j in range(data.shape[0])])

# observed data
data_z = -data[2,:] + noise[2,:]

def cost_function(params):
    m0 = ForwardMogiFactory(params[0:4])
    m1 = ForwardMogiFactory(params[4:])
    zdisp = filter_for_zdisp(m0(stations) + m1(stations))
    x = zdisp - data_z
    return 100000. * numpy.sum(real((conjugate(x)*x)))

def plot_noisy_data():
    import matplotlib.pyplot as plt
    plt.plot(stations[0,:],-data[2,:]+noise[2,:],'k.')
    plt.draw()
    plt.pause(0.001)

def plot_sol(params,linestyle='b-'):
    import matplotlib.pyplot as plt
    s0 = ForwardMogiFactory(params[0:4])
    s1 = ForwardMogiFactory(params[4:])
    xx = arange(-500,500,1)+1250.
    yy = 0*xx - 200.
    ss  = array((xx, yy))
    dd = s0(ss) + s1(ss)
    plt.plot(ss[0,:],-dd[2,:],'%s'%linestyle,linewidth=2.0)
    plt.draw()
    plt.pause(0.001)

ND = 8
NP = 80
MAX_GENERATIONS = 5000

def de_solve():
    solver = DifferentialEvolutionSolver(ND, NP)

    solver.enable_signal_handler()

    stepmon = VerboseMonitor()
    minrange = [-1000., -1000., -100., -1.]*2;
    maxrange = [1000., 1000., 100., 1.]*2;
    solver.SetRandomInitialPoints(min = minrange, max = maxrange)
    solver.SetEvaluationLimits(generations=MAX_GENERATIONS)
    solver.SetGenerationMonitor(stepmon)

    solver.Solve(cost_function, \
                 termination=ChangeOverGeneration(generations=300), \
                 CrossProbability=0.5, ScalingFactor=0.5, \
                 sigint_callback = plot_sol)

    solution = solver.Solution()
  
    return solution, stepmon

if __name__ == '__main__':

    plt.ion()
    plot_noisy_data()
    desol, dstepmon = de_solve()
    print("desol: %s" % desol)
   #plot_sol(dstepmon.x[-100],'k-')

    plot_sol(desol,'r-')

    getch()

# end of file



#!/usr/bin/env python
#
# Author: Alta Fang (altafang @caltech and alta @princeton)
# Author: Mike McKerns (mmckerns @caltech and @uqfoundation)
# Copyright (c) 1997-2016 California Institute of Technology.
# Copyright (c) 2016-2022 The Uncertainty Quantification Foundation.
# License: 3-clause BSD.  The full license text is available at:
#  - https://github.com/uqfoundation/mystic/blob/master/LICENSE
"""test termination conditions. (defaults listed below)

VTR(tolerance = 0.005)
ChangeOverGeneration(tolerance = 1e-6, generations = 30)
NormalizedChangeOverGeneration(tolerance = 1e-4, generations = 10)
CandidateRelativeTolerance(xtol=1e-4, ftol=1e-4)
SolutionImprovement(tolerance = 1e-5)
NormalizedCostTarget(fval = None, tolerance = 1e-6, generations = 30)
VTRChangeOverGeneration(ftol = 0.005, gtol = 1e-6, generations = 30)
PopulationSpread(tolerance=1e-6)
GradientNormTolerance(tolerance=1e-5, norm=Inf)
"""
from mystic.termination import *
from numpy import inf

def test_terminators(test, func=lambda x:x[0], info=False, verbose=False, disp=False):
  res = test(lambda x,y:"", func, info, verbose)  #XXX: just check settings
  if disp: print(res) 
  else: assert bool(res) in (True, False)
  res = test(VTR(), func, info)
  if disp: print("VTR():%s" % res)
  else: assert bool(res) in (True, False)
  res = test(VTR(inf), func, info)
  if disp: print("VTR(inf):%s" % res)
  else: assert bool(res) in (True, False)
  res = test(ChangeOverGeneration(), func, info)
  if disp: print("COG():%s" %  res)
  else: assert bool(res) in (True, False)
  res = test(ChangeOverGeneration(generations=5), func, info)
  if disp: print("COG(gen=5):%s" % res)
  else: assert bool(res) in (True, False)
  res = test(NormalizedChangeOverGeneration(), func, info)
  if disp: print("NCOG():%s" % res)
  else: assert bool(res) in (True, False)
  res = test(NormalizedChangeOverGeneration(generations=5), func, info)
  if disp: print("NCOG(gen=5):%s" % res)
  else: assert bool(res) in (True, False)
  res = test(CandidateRelativeTolerance(), func, info)
  if disp: print("CTR():%s" % res)
  else: assert bool(res) in (True, False)
  res = test(CandidateRelativeTolerance(ftol=inf), func, info)
  if disp: print("CTR(ftol=inf):%s" % res)
  else: assert bool(res) in (True, False)
  res = test(CandidateRelativeTolerance(inf), func, info)
  if disp: print("CTR(inf):%s" % res)
  else: assert bool(res) in (True, False)
  res = test(SolutionImprovement(), func, info)
  if disp: print("SI():%s" % res)
  else: assert bool(res) in (True, False)
  res = test(SolutionImprovement(inf), func, info)
  if disp: print("SI(inf):%s" % res)
  else: assert bool(res) in (True, False)
  res = test(NormalizedCostTarget(), func, info)
  if disp: print("NCT():%s" % res)
  else: assert bool(res) in (True, False)
  res = test(NormalizedCostTarget(generations=5), func, info)
  if disp: print("NCT(gen=5):%s" % res)
  else: assert bool(res) in (True, False)
  res = test(NormalizedCostTarget(generations=None), func, info)
  if disp: print("NCT(gen=None):%s" % res)
  else: assert bool(res) in (True, False)
  res = test(NormalizedCostTarget(inf,inf), func, info)
  if disp: print("NCT(inf,inf):%s" % res)
  else: assert bool(res) in (True, False)
  res = test(VTRChangeOverGeneration(), func, info)
  if disp: print("VCOG():%s" % res)
  else: assert bool(res) in (True, False)
  res = test(VTRChangeOverGeneration(generations=5), func, info)
  if disp: print("VCOG(gen=5):%s" % res)
  else: assert bool(res) in (True, False)
  res = test(VTRChangeOverGeneration(inf), func, info)
  if disp: print("VCOG(inf):%s" % res)
  else: assert bool(res) in (True, False)
  res = test(PopulationSpread(), func, info)
  if disp: print("PS():%s" % res)
  else: assert bool(res) in (True, False)
  res = test(PopulationSpread(inf), func, info)
  if disp: print("PS(inf):%s" % res)
  else: assert bool(res) in (True, False)
 #print("GNT():%s" % test(GradientNormTolerance(), func, info))
  return

def verbosity(solver):
    print("energy_history:%s" % solver.energy_history)
    print("population:%s" % solver.population)
    print("popEnergy:%s" % solver.popEnergy)
    print("bestSolution:%s" % solver.bestSolution)
    print("trialSolution:%s" % solver.trialSolution)
    print("bestEnergy:%s" % solver.bestEnergy)
    return

def test01(terminate, func=lambda x:x[0], info=False, debug=False):
  from mystic.solvers import DifferentialEvolutionSolver2 as DE2
  solver = DE2(3,5)
  solver.SetRandomInitialPoints()
  solver.SetEvaluationLimits(8)
  solver.Solve(func, VTR())
  if debug: verbosity(solver)
  return terminate(solver, info)

def test02(terminate, func=lambda x:x[0], info=False, debug=False):
  from mystic.solvers import DifferentialEvolutionSolver2 as DE2
 #solver = DE2(3,1) #Solver throws ValueError "sample larger than population"
 #solver = DE2(1,1) #Solver throws ValueError "sample larger than population"
  solver = DE2(1,5)
  solver.SetRandomInitialPoints()
  solver.SetEvaluationLimits(8)
  solver.Solve(func, VTR())
  if debug: verbosity(solver)
  return terminate(solver, info)

def test03(terminate, func=lambda x:x[0], info=False, debug=False):
  from mystic.solvers import DifferentialEvolutionSolver as DE
  solver = DE(3,5)
  solver.SetRandomInitialPoints()
  solver.SetEvaluationLimits(8)
  solver.Solve(func, VTR())
  if debug: verbosity(solver)
  return terminate(solver, info)

def test04(terminate, func=lambda x:x[0], info=False, debug=False):
  from mystic.solvers import DifferentialEvolutionSolver as DE
  solver = DE(1,5)
  solver.SetRandomInitialPoints()
  solver.SetEvaluationLimits(8)
  solver.Solve(func, VTR())
  if debug: verbosity(solver)
  return terminate(solver, info)

def test05(terminate, func=lambda x:x[0], info=False, debug=False):
  from mystic.solvers import NelderMeadSimplexSolver as NM
  solver = NM(3)
  solver.SetRandomInitialPoints()
  solver.SetEvaluationLimits(8)
  solver.Solve(func, VTR())
  if debug: verbosity(solver)
  return terminate(solver, info)

def test06(terminate, func=lambda x:x[0], info=False, debug=False):
  from mystic.solvers import NelderMeadSimplexSolver as NM
  solver = NM(1)
  solver.SetRandomInitialPoints()
  solver.SetEvaluationLimits(8)
  solver.Solve(func, VTR())
  if debug: verbosity(solver)
  return terminate(solver, info)

def test07(terminate, func=lambda x:x[0], info=False, debug=False):
  from mystic.solvers import PowellDirectionalSolver as PDS
  solver = PDS(3)
  solver.SetRandomInitialPoints()
  solver.SetEvaluationLimits(8)
  solver.Solve(func, VTR())
  if debug: verbosity(solver)
  return terminate(solver, info)

def test08(terminate, func=lambda x:x[0], info=False, debug=False):
  from mystic.solvers import PowellDirectionalSolver as PDS
  solver = PDS(1)
  solver.SetRandomInitialPoints()
  solver.SetEvaluationLimits(8)
  solver.Solve(func, VTR())
  if debug: verbosity(solver)
  return terminate(solver, info)


if __name__ == "__main__":
  disp = False
  verbose = False
  info = False
  """NOTES: For x:x[0], test01-test04 returns either lists or floats;
while test05-test06 returns a ndarray for population, popEnergy, bestSolution;
test07-test08 throw a RuntimeError "Too many iterations" due to "bracket()".
  For x:inf, test01-test02 have 1e+20 in energy_history, popEnergy, bestEnergy;
while test03-test04 have inf; test05-test06 have Inf in popEnergy;
while test07-test08 have array(inf) for energy_history, popEnergy,
and inf for bestEnergy, and energy_history, popEnergy, population is list
of ndarrays, while bestSolution is an ndarray.
  For x:10.0, test01-test06 notes are as for x:x[0]; while test07-test08
notes are same as x:inf except with 'inf' replaced by '10.0'.
  For x:0.0, x:-10.0, x:-inf, test01-test08 notes are as for x:10.0 except
with '10.0' replaced by '0.0', '-10.0', '-inf' respectively.

ISSUES:
  Many of the 'internals' do not have a standard format. Is this a problem?
  Why does x:inf result in 1e+20 for DE2?
  """
 #function = lambda x:x[0]
 #function = lambda x:inf
 #function = lambda x:10.0
  function = lambda x:0.0
 #function = lambda x:-10.0
 #function = lambda x:-inf

 #test_terminators(test01,function,info,verbose,disp)
 #test_terminators(test02,function,info,verbose,disp)
 #test_terminators(test03,function,info,verbose,disp)
 #test_terminators(test04,function,info,verbose,disp)
 #test_terminators(test05,function,info,verbose,disp)
 #test_terminators(test06,function,info,verbose,disp)
  test_terminators(test07,function,info,verbose,disp)
 #test_terminators(test08,function,info,verbose,disp)

# EOF



#!/usr/bin/env python
#
# Author: Mike McKerns (mmckerns @caltech and @uqfoundation)
# Copyright (c) 1997-2016 California Institute of Technology.
# Copyright (c) 2016-2022 The Uncertainty Quantification Foundation.
# License: 3-clause BSD.  The full license text is available at:
#  - https://github.com/uqfoundation/mystic/blob/master/LICENSE

import mystic.collapse as ct
import numpy as np
import mystic as my
m = my.monitors._load('_log.py')
# cleanup *pyc
import os
try: os.remove('_log.pyc')
except OSError: pass

# at
x = ct.collapse_at(m, target=0.0, tolerance=0.05, mask=None)
assert x == set((1, 7, 8, 9, 10, 11, 12, 14))
x = ct.collapse_at(m, target=0.0, tolerance=0.05, mask=set((10,11,12)))
assert x == set((1, 7, 8, 9, 14))
try: ct.collapse_at(m, target=0.0, mask=10); raise RuntimeError()
except TypeError: pass
try: ct.collapse_at(m, target=0.0, mask=(10,11,12)); raise RuntimeError()
except TypeError: pass
try: ct.collapse_at(m, target=0.0, mask=[(10,11,12)])
except TypeError: pass
try: ct.collapse_at(m, target=0.0, mask=[(10,11,12),(13,14,15)]); raise RuntimeError()
except TypeError: pass
try: ct.collapse_at(m, target=0.0, mask=((10,11),(12,13))); raise RuntimeError()
except TypeError: pass
try: ct.collapse_at(m, target=0.0, mask=(np.array((10,11)),np.array((12,13)))); raise RuntimeError()
except TypeError: pass
try: ct.collapse_at(m, target=0.0, mask=set(((10,11),(12,13)))); raise RuntimeError()
except ValueError: pass
try: ct.collapse_at(m, target=0.0, mask={(10,11):(12,13)}); raise RuntimeError()
except TypeError: pass


# as
x = ct.collapse_as(m, tolerance=0.05, mask=None)
assert x == set(((10, 11), (7, 12), (10, 12), (8, 9), (11, 14), (7, 11), (1, 11), (16, 17), (8, 14), (1, 14), (8, 10), (9, 11), (7, 10), (1, 10), (7, 14), (9, 14), (12, 14), (8, 11), (9, 10), (1, 9), (11, 12), (7, 9), (1, 12), (8, 12), (3, 4), (1, 8), (10, 14), (6, 13), (1, 7), (7, 8), (9, 12)))
x = ct.collapse_as(m, tolerance=0.05, mask=set(((10,11),(7,12),(10,12))))
assert x == set(((8, 9), (11, 14), (7, 11), (1, 11), (16, 17), (8, 14), (1, 14), (8, 10), (9, 11), (7, 10), (1, 10), (7, 14), (9, 14), (12, 14), (8, 11), (9, 10), (1, 9), (11, 12), (7, 9), (1, 12), (8, 12), (3, 4), (1, 8), (10, 14), (6, 13), (1, 7), (7, 8), (9, 12)))
x = ct.collapse_as(m, tolerance=0.05, mask=set((10,12)))
assert x == set(((8, 9), (11, 14), (7, 11), (1, 11), (16, 17), (8, 14), (1, 14), (9, 11), (7, 14), (9, 14), (8, 11), (1, 9), (7, 9), (3, 4), (1, 8), (6, 13), (1, 7), (7, 8)))
x = ct.collapse_as(m, tolerance=0.05, mask=set((10,12,(9,8),(14,11))))
assert x == set(((7, 11), (1, 11), (16, 17), (8, 14), (1, 14), (9, 11), (7, 14), (9, 14), (8, 11), (1, 9), (7, 9), (3, 4), (1, 8), (6, 13), (1, 7), (7, 8)))
try: ct.collapse_as(m, mask=10); raise RuntimeError()
except TypeError: pass
try: ct.collapse_as(m, mask=(10,11)); raise RuntimeError()
except TypeError: pass
try: ct.collapse_as(m, mask=(10,11,12,13)); raise RuntimeError()
except TypeError: pass
try: ct.collapse_as(m, mask=set(((10,11,12),(14,8,9)))); raise RuntimeError()
except ValueError: pass
try: ct.collapse_as(m, mask=[(10,11,12,13)]); raise RuntimeError()
except TypeError: pass
try: ct.collapse_as(m, mask=[(11,12),(14,9)]); raise RuntimeError()
except TypeError: pass
try: ct.collapse_as(m, mask=(np.array([11,12]),np.array([14,9]))); raise RuntimeError()
except TypeError: pass
try: ct.collapse_as(m, mask=(np.array([10,11,12]),np.array([14,8,9]))); raise RuntimeError()
except TypeError: pass
try: ct.collapse_as(m, mask={11:9}); raise RuntimeError()
except TypeError: pass
try: ct.collapse_as(m, mask={(11,12):(14,9)}); raise RuntimeError()
except TypeError: pass


# weights
x = ct.collapse_weight(m, mask=None)
assert x == {0:set((1,)), 1:set((1,2)), 2:set((0,2))}
x = ct.collapse_weight(m, mask={0:set((1,)), 1:set((0,1))})
assert x == {1:set((2,)), 2:set((0,2))}
x = ct.collapse_weight(m, mask=((0,1),(1,2)))
assert x == ((1,2,2),(1,0,2))
x = ct.collapse_weight(m, mask=set(((0,1),(1,2),(2,0))))
assert x == set(((1,1),(2,2)))
try: ct.collapse_weight(m, mask=(None,None)); raise RuntimeError()
except TypeError: pass
try: ct.collapse_weight(m, mask=(0,None)); raise RuntimeError()
except TypeError: pass
try: ct.collapse_weight(m, mask=(None,2)); raise RuntimeError()
except TypeError: pass
try: ct.collapse_weight(m, mask=(1,2)); raise RuntimeError()
except TypeError: pass
try: ct.collapse_weight(m, mask=((1,2),None)); raise RuntimeError()
except TypeError: pass
try: ct.collapse_weight(m, mask=(None,(1,2))); raise RuntimeError()
except TypeError: pass
x = ct.collapse_weight(m, mask=((0,1,2),(0,1,2)))
assert x == ((0, 1, 2), (1, 2, 0))
try: ct.collapse_weight(m, mask=(slice(1,None,1),slice(1,None,1))); raise RuntimeError()
except TypeError: pass
try: ct.collapse_weight(m, mask=(slice(None,2,1),slice(None,2,1))); raise RuntimeError()
except TypeError: pass
try: ct.collapse_weight(m, mask=(None,set((0,1)))); raise RuntimeError()
except TypeError: pass
try: ct.collapse_weight(m, mask=(None,set(((0,1))))); raise RuntimeError()
except TypeError: pass
try: ct.collapse_weight(m, mask=(None,{0:1})); raise RuntimeError()
except TypeError: pass
try: ct.collapse_weight(m, mask=(None,{0:(1,)})); raise RuntimeError()
except TypeError: pass
try: ct.collapse_weight(m, mask=(None,[(1,2)])); raise RuntimeError()
except TypeError: pass
try: ct.collapse_weight(m, mask=(None,[(0,1),(1,2)])); raise RuntimeError()
except TypeError: pass
try: ct.collapse_weight(m, mask=([(1,2)],[(1,2)])); raise RuntimeError()
except TypeError: pass
try: ct.collapse_weight(m, mask=([(0,1),(1,2)],[(0,1),(1,2)])); raise RuntimeError()
except TypeError: pass
try: ct.collapse_weight(m, mask=set((None,))); raise RuntimeError()
except ValueError: pass
try: ct.collapse_weight(m, mask=set((((0,1),(1,2)),))); raise RuntimeError()
except ValueError: pass
try: ct.collapse_weight(m, mask=set((((0,1),(1,2)),((1,2),(2,2))))); raise RuntimeError()
except ValueError: pass
try: ct.collapse_weight(m, mask=set((0,))); raise RuntimeError() # error
except ValueError: pass
try: ct.collapse_weight(m, mask=set(((0,)))); raise RuntimeError()
except ValueError: pass
try: ct.collapse_weight(m, mask=set(((0,1,2)))); raise RuntimeError()
except ValueError: pass
try: ct.collapse_weight(m, mask=set((((0,1,1,2))))); raise RuntimeError()
except ValueError: pass
try: ct.collapse_weight(m, mask=set(([(0,1)],[(1,2)]))); raise RuntimeError()
except TypeError: pass
try: ct.collapse_weight(m, mask=set((((0,1),),((1,2),)))); raise RuntimeError()
except ValueError: pass
try: ct.collapse_weight(m, mask={None:None}); raise RuntimeError()
except ValueError: pass
try: ct.collapse_weight(m, mask={1:None}); raise RuntimeError()
except ValueError: pass
try: ct.collapse_weight(m, mask={None:2}); raise RuntimeError()
except ValueError: pass
try: ct.collapse_weight(m, mask={1:2}); raise RuntimeError()
except ValueError: pass
try: ct.collapse_weight(m, mask={None:(1,)}); raise RuntimeError()
except ValueError: pass
try: ct.collapse_weight(m, mask={1:(1,)}); raise RuntimeError()
except ValueError: pass
try: ct.collapse_weight(m, mask={0:(0,1,2)}); raise RuntimeError()
except ValueError: pass
x = ct.collapse_weight(m, mask={0:set((0,1,2))})
assert x == {1:set((1,2)), 2:set((0,2))}
try: ct.collapse_weight(m, mask={0:set(((0,1,2),))}); raise RuntimeError()
except ValueError: pass
try: ct.collapse_weight(m, mask={0:[(0,1,2)]}); raise RuntimeError()
except ValueError: pass
try: ct.collapse_weight(m, mask={0:[(0,1),(2,0)]}); raise RuntimeError()
except ValueError: pass
try: ct.collapse_weight(m, mask={None:[(0,1),(2,0)]}); raise RuntimeError()
except ValueError: pass
try: ct.collapse_weight(m, mask={(0,1):[(0,1),(2,0)]}); raise RuntimeError()
except ValueError: pass
try: ct.collapse_weight(m, mask={(1,):(1,)}); raise RuntimeError()
except ValueError: pass
try: ct.collapse_weight(m, mask={(1,2):(1,2)}); raise RuntimeError()
except ValueError: pass
try: ct.collapse_weight(m, mask={(1,0):(1,2)}); raise RuntimeError()
except ValueError: pass
try: ct.collapse_weight(m, mask={(0,1):(1,2)}); raise RuntimeError()
except ValueError: pass
try: ct.collapse_weight(m, mask={(0,1,2):(0,1,2)}); raise RuntimeError()
except ValueError: pass


# positions
x = ct.collapse_position(m, mask=None)
assert x == {0:set(((0,1),)), 1:set(((0,1),(1,2),(0,2)),), 2:set(((1,2),))}
x = ct.collapse_position(m, mask={0:set(((0,1),)), 1:set(((1,2),))})
assert x == {1:set(((0,1),(0,2)),), 2:set(((1,2),))}
x = ct.collapse_position(m, mask=((0,1,2),((0,1),(1,2),(0,2))))
assert x == ((1,1,2), ((0,1),(0,2),(1,2)))
x = ct.collapse_position(m, mask=set(((0,(0,1)),(1,(1,2)),(2,(2,0)))))
assert x == set(((1,(0,1)),(1,(0,2)),(2,(1,2))))
try: ct.collapse_position(m, mask=(None,None)); raise RuntimeError()
except TypeError: pass
try: ct.collapse_position(m, mask=(0,None)); raise RuntimeError()
except TypeError: pass
try: ct.collapse_position(m, mask=(None,2)); raise RuntimeError()
except TypeError: pass
try: ct.collapse_position(m, mask=(1,2)); raise RuntimeError()
except TypeError: pass
try: ct.collapse_position(m, mask=((1,2),None)); raise RuntimeError()
except TypeError: pass
try: ct.collapse_position(m, mask=(None,(1,2))); raise RuntimeError()
except TypeError: pass
try: ct.collapse_position(m, mask=((1,2),(1,2))); raise RuntimeError()
except TypeError: pass
try: ct.collapse_position(m, mask=((0,1),(0,1))); raise RuntimeError()
except TypeError: pass
try: ct.collapse_position(m, mask=((0,1,2),(0,1,2))); raise RuntimeError()
except TypeError: pass
try: ct.collapse_position(m, mask=(slice(1,None,1),slice(1,None,1))); raise RuntimeError()
except TypeError: pass
try: ct.collapse_position(m, mask=(slice(None,2,1),slice(None,2,1))); raise RuntimeError()
except TypeError: pass
try: ct.collapse_position(m, mask=(None,set((0,1)))); raise RuntimeError() # error
except TypeError: pass
try: ct.collapse_position(m, mask=(None,set(((0,1))))); raise RuntimeError()
except TypeError: pass
try: ct.collapse_position(m, mask=(None,{0:1})); raise RuntimeError()
except TypeError: pass
try: ct.collapse_position(m, mask=(None,{0:(1,)})); raise RuntimeError()
except TypeError: pass
try: ct.collapse_position(m, mask=(None,[(1,2)])); raise RuntimeError()
except TypeError: pass
try: ct.collapse_position(m, mask=(None,[(0,1),(1,2)])); raise RuntimeError()
except TypeError: pass
##### 4 CASES #####
x = ct.collapse_position(m, mask=((0,1,1), [(0,1),(0,1),(1,2)]))
assert x == ((1,2), ((0,2),(1,2)))
x = ct.collapse_position(m, mask=set(((0,(0,1)), (1,(0,1)), (1,(1,2)))))
assert x == set(((1,(0,2)), (2,(1,2))))
x = ct.collapse_position(m, mask={0:set(((0,1),)), 1:set(((0,1),(1,2)))})
assert x == {1:set(((0,2),)), 2:set(((1,2),))}
try: ct.collapse_position(m, mask={0:[(0,1)], 1:[(0,1),(1,2)]}); raise RuntimeError()
except ValueError: pass
###################
try: ct.collapse_position(m, mask=([(1,2)],[(1,2)])); raise RuntimeError()
except TypeError: pass
try: ct.collapse_position(m, mask=([(0,1),(1,2)],[(0,1),(1,2)])); raise RuntimeError()
except TypeError: pass
try: ct.collapse_position(m, mask=set((None,))); raise RuntimeError()
except ValueError: pass
try: ct.collapse_position(m, mask=set(((0,1)))); raise RuntimeError()
except ValueError: pass
try: ct.collapse_position(m, mask=set(((0,1),(1,2)))); raise RuntimeError()
except ValueError: pass
try: ct.collapse_position(m, mask=set((((0,1),(1,2))))); raise RuntimeError() # error
except ValueError: pass
try: ct.collapse_position(m, mask=set((((0,1),(1,2)),((1,2),(2,2))))); raise RuntimeError()
except ValueError: pass
try: ct.collapse_position(m, mask=set((0,))); raise RuntimeError()
except ValueError: pass
try: ct.collapse_position(m, mask=set(([(0,1)],[(1,2)]))); raise RuntimeError()
except TypeError: pass
try: ct.collapse_position(m, mask=set((((0,1),),((1,2),)))); raise RuntimeError()
except ValueError: pass
try: ct.collapse_position(m, mask={1:2})
except ValueError: pass
try: ct.collapse_position(m, mask={None:(1,)})
except ValueError: pass
try: ct.collapse_position(m, mask={1:(1,)})
except ValueError: pass
try: ct.collapse_position(m, mask={(1,):(1,)})
except ValueError: pass
try: ct.collapse_position(m, mask={0:set((0,1,2))}); raise RuntimeError()
except ValueError: pass
try: ct.collapse_position(m, mask={0:set(((0,1,2)))})
except ValueError: pass
try: ct.collapse_position(m, mask={0:[(0,1,2)]})
except ValueError: pass
try: ct.collapse_position(m, mask=set(((0,)))); raise RuntimeError()  # <wrong size>
except ValueError: pass
try: ct.collapse_position(m, mask=set(((0,1,2)))); raise RuntimeError()
except ValueError: pass
try: ct.collapse_position(m, mask=set((((0,1,1,2))))); raise RuntimeError()
except ValueError: pass
try: ct.collapse_position(m, mask={None:None})
except ValueError: pass
try: ct.collapse_position(m, mask={1:None}) # any
except ValueError: pass
try: ct.collapse_position(m, mask={None:2})
except ValueError: pass
try: ct.collapse_position(m, mask={0:(0,1,2)}) # only
except ValueError: pass
try: ct.collapse_position(m, mask={0:[(0,1),(1,2)]}) # <numpy>
except ValueError: pass
try: ct.collapse_position(m, mask={None:[(0,1),(1,2)]})
except ValueError: pass
try: ct.collapse_position(m, mask={(0,1):[(0,1)]})
except ValueError: pass
try: ct.collapse_position(m, mask={(1,2):[(1,2)]})
except ValueError: pass
try: ct.collapse_position(m, mask={(0,1):[(0,1),(1,2)]})
except ValueError: pass
try: ct.collapse_position(m, mask={(1,2):(1,2)})
except ValueError: pass
try: ct.collapse_position(m, mask={(1,0):(1,2)})
except ValueError: pass
try: ct.collapse_position(m, mask={(0,1):(1,2)})
except ValueError: pass
try: ct.collapse_position(m, mask={(0,1):(0,1)})
except ValueError: pass
try: ct.collapse_position(m, mask={(0,1,2):(0,1,2)})
except ValueError: pass



# EOF



#!/usr/bin/env python
#
# Author: Mike McKerns (mmckerns @caltech and @uqfoundation)
# Copyright (c) 1997-2016 California Institute of Technology.
# Copyright (c) 2016-2022 The Uncertainty Quantification Foundation.
# License: 3-clause BSD.  The full license text is available at:
#  - https://github.com/uqfoundation/mystic/blob/master/LICENSE

from mystic.coupler import *
from mystic.math import almostEqual

def test_outer():

  def squared(x):
    return x**2

  @outer(outer=squared) 
  def plus_one_squared(x):
    return x+1

  from numpy import array
  x = array([1,2,3,4,5])
  assert all(plus_one_squared(x)) == all((x+1)**2)


def test_inner():

  def squared(x):
    return x**2

  @inner(inner=squared) 
  def squared_plus_one(x):
    return x+1

  from numpy import array
  x = array([1,2,3,4,5])
  assert all(squared_plus_one(x)) == all(x**2 + 1)


def test_outer_constraint():

  from mystic.math.measures import impose_mean, mean

  def impose_constraints(x, mean, weights=None):
    return impose_mean(mean, x, weights)

  @outer(outer=impose_constraints, kwds={'mean':5.0})
  def mean_of_squared(x):
    return [i**2 for i in x]

  from numpy import array
  x = array([1,2,3,4,5])
  y = impose_mean(5, [i**2 for i in x])
  assert mean(y) == 5.0
  assert mean_of_squared(x) == y


def test_inner_constraint():

  from mystic.math.measures import impose_mean

  def impose_constraints(x, mean, weights=None):
    return impose_mean(mean, x, weights)

  @inner(inner=impose_constraints, kwds={'mean':5.0})
  def mean_then_squared(x):
    return [i**2 for i in x]

  from numpy import array
  x = array([1,2,3,4,5])
  assert mean_then_squared(x) == [i**2 for i in impose_mean(5,x)]


def test_proxified_constraint():

  from mystic.math.measures import impose_mean

  @inner_proxy(inner=impose_mean)
  def mean_then_squared(x): #XXX: proxy doesn't preserve function signature
    return [i**2 for i in x]

  from numpy import array
  x = array([1,2,3,4,5])
  assert mean_then_squared(5,x) == [i**2 for i in impose_mean(5,x)]


def test_inner_constraints():

  from mystic.math.measures import impose_mean, impose_spread

  def impose_constraints(x, mean=0.0, spread=1.0):
    x = impose_mean(mean, x)
    x = impose_spread(spread, x)
    return x

  @inner(inner=impose_constraints, kwds={'mean':5.0, 'spread':50.0})
  def constrained_squared(x):
    return [i**2 for i in x]

  from numpy import array
  x = array([1,2,3,4,5])
  y = impose_spread(50.0, impose_mean(5.0,x))
  assert constrained_squared(x) == [i**2 for i in y]


def test_proxified_constraints():

  from mystic.math.measures import impose_mean, impose_spread

  def impose_constraints(x, mean=0.0, spread=1.0):
    x = impose_mean(mean, x)
    x = impose_spread(spread, x)
    return x

  @inner_proxy(inner=impose_constraints)
  def constrained_squared(x): #XXX: proxy doesn't preserve function signature
    return [i**2 for i in x]

  from numpy import array
  x = array([1,2,3,4,5])
  y = impose_spread(50.0, impose_mean(5.0,x))
  assert constrained_squared(x, 5.0, 50.0) == [i**2 for i in y]


def test_constrain():

  from mystic.math.measures import mean, spread
  from mystic.math.measures import impose_mean, impose_spread
  def mean_constraint(x, mean=0.0):
    return impose_mean(mean, x)

  def range_constraint(x, spread=1.0):
    return impose_spread(spread, x)

  @inner(inner=range_constraint, kwds={'spread':5.0})
  @inner(inner=mean_constraint, kwds={'mean':5.0})
  def constraints(x):
    return x

  def cost(x):
    return abs(sum(x) - 5.0)

  from mystic.solvers import fmin_powell
  from numpy import array
  x = array([1,2,3,4,5])
  y = fmin_powell(cost, x, constraints=constraints, disp=False)

  assert mean(y) == 5.0
  assert spread(y) == 5.0
  assert almostEqual(cost(y), 4*(5.0))


if __name__ == '__main__':
  test_outer()
  test_inner()
  test_outer_constraint()
  test_inner_constraint()
  test_proxified_constraint()
  test_inner_constraints()
  test_proxified_constraints()
  test_constrain()


# EOF



#!/usr/bin/env python
#
# almostEqual is a repackaging of numpy.allclose, but at different 'tol'
# approx_equal is similar to almostEqual, and can be treated as deprecated
#
# Forked by: Mike McKerns (May 2010)
# Author: Mike McKerns (mmckerns @caltech and @uqfoundation)
# Copyright (c) 2010-2016 California Institute of Technology.
# Copyright (c) 2016-2022 The Uncertainty Quantification Foundation.
# License: 3-clause BSD.  The full license text is available at:
#  - https://github.com/uqfoundation/mystic/blob/master/LICENSE
"""
tools for measuring equality
"""

def _float_approx_equal(x, y, tol=1e-18, rel=1e-7):
    if tol is rel is None:
        raise TypeError('cannot specify both absolute and relative errors are None')
    tests = []
    if tol is not None: tests.append(tol)
    if rel is not None: tests.append(rel*abs(x))
    assert tests
    return abs(x - y) <= max(tests)


def approx_equal(x, y, *args, **kwargs):
    """Return True if x and y are approximately equal, otherwise False.

Args:
    x (object): first object to compare
    y (object): second object to compare
    tol (float, default=1e-18): absolute error
    rel (float, default=1e-7): relative error

Returns:
    True if x and y are equal within tolerance, otherwise returns False.

Notes:
    If x and y are floats, return True if y is within either absolute error
    tol or relative error rel of x. You can disable either the absolute or
    relative check by passing ``None`` as *tol* or *rel* (but not both).

    For any other objects, x and y are checked in that order for a method
    ``__approx_equal__``, and the result of that is returned as a bool. Any
    optional arguments are passed to the ``__approx_equal__`` method.

    ``__approx_equal__`` can return ``NotImplemented`` to signal it doesn't
    know how to perform the specific comparison, in which case the other
    object is checked instead. If neither object has the method, or both defer
    by returning ``NotImplemented``, then fall back on the same numeric
    comparison that is used for floats.

Examples:
    >>> approx_equal(1.2345678, 1.2345677)
    True
    >>> approx_equal(1.234, 1.235)
    False
    """
    if not (type(x) is type(y) is float):
        # Skip checking for __approx_equal__ in the common case of two floats.
        methodname = '__approx_equal__'
        # Allow the objects to specify what they consider "approximately equal",
        # giving precedence to x. If either object has the appropriate method, we
        # pass on any optional arguments untouched.
        for a,b in ((x, y), (y, x)):
            try:
                method = getattr(a, methodname)
            except AttributeError:
                continue
            else:
                result = method(b, *args, **kwargs)
                if result is NotImplemented:
                    continue
                return bool(result)
    # If we get here without returning, then neither x nor y knows how to do an
    # approximate equal comparison (or are both floats). Fall back to a numeric
    # comparison.
    return _float_approx_equal(x, y, *args, **kwargs)


def almostEqual(x, y, tol=1e-18, rel=1e-7):
    """
    Returns True if two arrays are element-wise equal within a tolerance.
    
    The tolerance values are positive, typically very small numbers.  The
    relative difference (`rtol` * abs(`b`)) and the absolute difference
    `atol` are added together to compare against the absolute difference
    between `a` and `b`.
    
    Parameters
    ----------
    a, b : array_like
        Input arrays to compare.
    rtol : float
        The relative tolerance parameter (see Notes).
    atol : float
        The absolute tolerance parameter (see Notes).
    
    Returns
    -------
    y : bool
        Returns True if the two arrays are equal within the given
        tolerance; False otherwise. If either array contains NaN, then
        False is returned.

    Notes
    -----
    If the following equation is element-wise True, then almostEqual returns
    True.
    
     absolute(`a` - `b`) <= (`atol` + `rtol` * absolute(`b`))
    
    Examples
    --------
    >>> almostEqual([1e10,1.2345678], [1e10,1.2345677])
    True
    >>> almostEqual([1e10,1.234], [1e10,1.235])
    False
"""
    from numpy import allclose
    return allclose(x, y, rtol=rel, atol=tol)


def tolerance(x, tol=1e-15, rel=1e-15):
    """relative plus absolute difference"""
    return tol + abs(x)*rel


# end of file



#!/usr/bin/env python
#
# Author: Patrick Hung (patrickh @caltech)
# Author: Mike McKerns (mmckerns @caltech and @uqfoundation)
# Copyright (c) 1997-2016 California Institute of Technology.
# Copyright (c) 2016-2022 The Uncertainty Quantification Foundation.
# License: 3-clause BSD.  The full license text is available at:
#  - https://github.com/uqfoundation/mystic/blob/master/LICENSE
"""
Similar to test_mogi2 (two sources) (See that one first)
"""

from test_mogi2 import params0, params1, stations, data, data_z, ND, NP, plot_sol, plot_noisy_data, MAX_GENERATIONS, ForwardMogiFactory
from mystic.solvers import DifferentialEvolutionSolver
from mystic.termination import ChangeOverGeneration, VTR
from mystic.monitors import Monitor
from mystic.tools import getch, random_seed

from mystic.forward_model import CostFactory
from mystic.filters import component

def de_solve(CF):
    solver = DifferentialEvolutionSolver(ND, NP)

    solver.enable_signal_handler()

    stepmon = Monitor()
    minrange = [-1000., -1000., -100., -1.]*2;
    maxrange = [1000., 1000., 100., 1.]*2;
    solver.SetRandomInitialPoints(min = minrange, max = maxrange)
    solver.SetEvaluationLimits(generations=MAX_GENERATIONS)
    solver.SetGenerationMonitor(stepmon)

    solver.Solve(CF, termination = ChangeOverGeneration(generations=300), \
                 CrossProbability=0.5, ScalingFactor=0.5, \
                 sigint_callback = plot_sol)

    solution = solver.Solution()
  
    return solution, stepmon

if __name__ == '__main__':

    F = CostFactory()
    F.addModel(ForwardMogiFactory, 4, 'mogi1', outputFilter=component(2, -1))
    F.addModel(ForwardMogiFactory, 4, 'mogi2', outputFilter=component(2, -1))
    myCostFunction = F.getCostFunction(evalpts=stations, observations=data_z)
    print(F)

    def C2(x):
        "This is the new version"
        return 100000 * myCostFunction(x)

    def C3(x):
        "Cost function constructed by hand"
        from test_mogi2 import cost_function
        return cost_function(x)

    def test():
        "call me to see if the functions return the same thing"
        rp = F.getRandomParams()
        print("C2: %s" % C2(rp))
        print("C3: %s" % C3(rp))

    test()
    import matplotlib.pyplot as plt
    plot_noisy_data()
    desol, dstepmon = de_solve(C2)
    print("desol: %s" % desol)

   #plot_sol(dstepmon.x[-100],'k-')
    plot_sol(desol,'r-')

    getch()

# end of file



#!/usr/bin/env python
#
# Author: Mike McKerns (mmckerns @caltech and @uqfoundation)
# Copyright (c) 1997-2016 California Institute of Technology.
# Copyright (c) 2016-2022 The Uncertainty Quantification Foundation.
# License: 3-clause BSD.  The full license text is available at:
#  - https://github.com/uqfoundation/mystic/blob/master/LICENSE
#
# Abstract Solver Class
"""
This module contains the base class for mystic solvers that utilize
a parallel ``map`` function to enable parallel computing.  This module
describes the map solver interface. As with the ``AbstractSolver``, the
``_Step`` method must be overwritten with the derived solver's optimization
algorithm. Additionally, for the ``AbstractMapSolver``, a call to ``map`` is
required. In addition to the class interface, a simple function interface for
a derived solver class is often provided. For an example, see the following.

The default map API settings are provided within mystic, while distributed
and parallel computing maps can be obtained from the ``pathos`` package 
(http://dev.danse.us/trac/pathos).

Examples:

    A typical call to a 'map' solver will roughly follow this example:

    >>> # the function to be minimized and the initial values
    >>> from mystic.models import rosen
    >>> lb = [0.0, 0.0, 0.0]
    >>> ub = [2.0, 2.0, 2.0]
    >>> 
    >>> # get monitors and termination condition objects
    >>> from mystic.monitors import Monitor
    >>> stepmon = Monitor()
    >>> from mystic.termination import CandidateRelativeTolerance as CRT
    >>> 
    >>> # select the parallel launch configuration
    >>> from pyina.launchers import Mpi as Pool
    >>> NNODES = 4
    >>> npts = 20
    >>>
    >>> # instantiate and configure the solver
    >>> from mystic.solvers import BuckshotSolver
    >>> solver = BuckshotSolver(len(lb), npts)
    >>> solver.SetMapper(Pool(NNODES).map)
    >>> solver.SetGenerationMonitor(stepmon)
    >>> solver.SetTermination(CRT())
    >>> solver.Solve(rosen)
    >>> 
    >>> # obtain the solution
    >>> solution = solver.Solution()


Handler
=======

All solvers packaged with mystic include a signal handler that provides
the following options::

    sol: Print current best solution.
    cont: Continue calculation.
    call: Executes sigint_callback, if provided.
    exit: Exits with current best solution.

Handlers are enabled with the ``enable_signal_handler`` method,
and are configured through the solver's ``Solve`` method.  Handlers
trigger when a signal interrupt (usually, ``Ctrl-C``) is given while
the solver is running.

Notes:

    The handler is currently disabled when the solver is run in parallel.
"""
__all__ = ['AbstractMapSolver']


from mystic.monitors import Null
from mystic.abstract_solver import AbstractSolver


class AbstractMapSolver(AbstractSolver):
    """
AbstractMapSolver base class for mystic optimizers that utilize parallel map.
    """

    def __init__(self, dim, **kwds):
        """
Takes one initial input::

    dim      -- dimensionality of the problem.

Additional inputs::

    npop     -- size of the trial solution population.       [default = 1]

Important class members::

    nDim, nPop       = dim, npop
    generations      - an iteration counter.
    evaluations      - an evaluation counter.
    bestEnergy       - current best energy.
    bestSolution     - current best parameter set.           [size = dim]
    popEnergy        - set of all trial energy solutions.    [size = npop]
    population       - set of all trial parameter solutions. [size = dim*npop]
    solution_history - history of bestSolution status.       [StepMonitor.x]
    energy_history   - history of bestEnergy status.         [StepMonitor.y]
    signal_handler   - catches the interrupt signal.         [***disabled***]
        """
        super(AbstractMapSolver, self).__init__(dim, **kwds)
       #AbstractSolver.__init__(self,dim,**kwds)
       #self.signal_handler   = None
       #self._handle_sigint   = False
        trialPop = [[0.0 for i in range(dim)] for j in range(self.nPop)]
        self.trialSolution    = trialPop
        self._map_solver      = True

        # import 'map' defaults
        from mystic.python_map import serial_launcher
        from mystic.python_map import python_map
        from mystic.python_map import worker_pool
        from mystic.python_map import defaults

        # default settings for parallel and distributed computing
        launcher        = serial_launcher   # launcher
        mapper          = worker_pool       # map_strategy
        nodes           = int(defaults['nodes'])
        scheduler       = defaults['scheduler'] # scheduler
        queue           = defaults['queue'] # scheduler_queue
        timelimit       = defaults['timelimit']
        servers         = ('*',) #<detect>  # hostname:port
        ncpus           = None   #<detect>  # local processors
       #servers         = ()     #<None>    # hostname:port
       #ncpus           = 0      #<None>    # local processors
        self._mapconfig = dict(nodes=nodes, launcher=launcher, \
                               mapper=mapper, queue=queue, \
                               timelimit=timelimit, scheduler=scheduler, \
                               ncpus=ncpus, servers=servers)
        self._map       = python_map        # map
        return

    def SelectServers(self, servers, ncpus=None): #XXX: needs some thought...
        """Select the compute server.

    Accepts a tuple of ('hostname:port',), listing each available
    computing server.

    If ncpus=None, then 'autodetect'; or if ncpus=0, then 'no local'.
    If servers=('*',), then 'autodetect'; or if servers=(), then 'no remote'.

Inputs:
    servers -- tuple of compute servers  [DEFAULT: autodetect]

Additional inputs:
    ncpus -- number of local processors  [DEFAULT: autodetect]
        """
        self._mapconfig['servers'] = servers
        self._mapconfig['ncpus'] = ncpus  #XXX: merge with nodes, somehow ???
       #print("known servers: %s" % str(servers))
       #print("known # of local processors: %s" % str(ncpus))
        return

    def SetMapper(self, map, strategy=None): #XXX: use strategy+format ?
        """Set the map function and the mapping strategy.

    Sets a mapping function to perform the map-reduce algorithm.
    Uses a mapping strategy to provide the algorithm for distributing
    the work list of optimization jobs across available resources.

Inputs:
    map -- the mapping function [DEFAULT: python_map]
    strategy -- map strategy (see pyina.mappers) [DEFAULT: worker_pool]
        """
        self._map = map
        if strategy:
          self._mapconfig['mapper'] = strategy
        #FIXME: not a true mapping function... just a dummy interface to a str
        # a real mapper has map(func,*args) interface... this expects map().
        # should be...
        #   def xxx_map( f, input, *args, **kwds ):
        #     ...
        #     from yyy.xxx import map
        #     result = map(f, input, *args, **kwds)
        #     ...
        #     return result
        return

    def SetLauncher(self, launcher, nnodes=None): #XXX: use run+scheduler ?
        """Set launcher and (optionally) number of nodes.

    Uses a launcher to provide the solver with the syntax to
    configure and launch optimization jobs on the selected resource.

Inputs:
    launcher -- launcher function (see pyina.launchers)  [DEFAULT: serial_launcher]

Additional inputs:
    nnodes -- number of parallel compute nodes  [DEFAULT: 1]
        """
       #XXX: should be a Launcher class, not a function (see pathos.SSHLauncher)
        self._mapconfig['launcher'] = launcher
       #if launcher != python_map: #FIXME: CANNOT currently change to ez_map!!!
       #    exec("import pyina")   #FIXME: launcher should provide launcher.map
       #    exec("self._map = pyina.ez_map")
        self._mapconfig['nodes'] = nnodes
        return

    def SelectScheduler(self, scheduler, queue, timelimit=None):
        """Select scheduler and queue (and optionally) timelimit.

    Takes a scheduler function and a string queue name to submit
    the optimization job. Additionally takes string time limit
    for scheduled job.

    Example: scheduler, queue='normal', timelimit='00:02'

Inputs:
    scheduler -- scheduler function (see pyina.launchers)  [DEFAULT: None]
    queue -- queue name string (see pyina.launchers)  [DEFAULT: None]

Additional inputs:
    timelimit -- time string HH:MM:SS format  [DEFAULT: '00:05:00']
        """
        self._mapconfig['scheduler'] = scheduler
        self._mapconfig['queue'] = queue
        self._mapconfig['timelimit'] = timelimit
        return


if __name__=='__main__':
    help(__name__)

# end of file



#!/usr/bin/env python
#
# Author: Mike McKerns (mmckerns @caltech and @uqfoundation)
# Copyright (c) 2009-2016 California Institute of Technology.
# Copyright (c) 2016-2022 The Uncertainty Quantification Foundation.
# License: 3-clause BSD.  The full license text is available at:
#  - https://github.com/uqfoundation/mystic/blob/master/LICENSE

from mystic.math.grid import gridpts

#######################################################################
# scaling and mpi info; also optimizer configuration parameters
# hard-wired: use fmin solver
#######################################################################
#scale = 1.0

#npop = 20
nbins = [2,2,2]
#maxiter = 1000
#maxfun = 1e+6
#convergence_tol = 1e-4

# MPI config
nnodes = '4:ppn=1,partition=hera'   # = npop

# QUEUE config
queue = 'pdebug'
timelimit = '00:10'


#######################################################################
# the model function
#######################################################################
#from surrogate import marc_surr as model
from surrogate import ballistic_limit as limit


#######################################################################
# the subdiameter calculation
#######################################################################
def costFactory(i):
  """a cost factory for the cost function"""

  def cost(rv):
    """compute the diameter as a calculation of cost

  Input:
    - rv -- 1-d array of model parameters

  Output:
    - diameter -- scale * | F(x) - F(x')|**2
    """
    from surrogate import marc_surr as model

    # prepare x and xprime
    rv = list(rv)
    params = rv[:-1]                         #XXX: assumes Xi' is at rv[-1]
    params_prime = rv[:i]+rv[-1:]+rv[i+1:-1] #XXX: assumes Xi' is at rv[-1]

    # get the F(x) response
    Fx = model(params)

    # get the F(x') response
    Fxp = model(params_prime)

    # compute diameter
    scale = 1.0
    return -scale * (Fx - Fxp)**2

  return cost


#######################################################################
# the steepest descent optimizer
#######################################################################
def local_optimize(cost,x0,lb,ub):
  from mystic.solvers import PowellDirectionalSolver
  from mystic.termination import NormalizedChangeOverGeneration as NCOG
  from mystic.monitors import VerboseMonitor, Monitor

  maxiter = 1000
  maxfun = 1e+6
  convergence_tol = 1e-4

 #stepmon = VerboseMonitor(100)
  stepmon = Monitor()
  evalmon = Monitor()

  ndim = len(lb)

  solver = PowellDirectionalSolver(ndim)
  solver.SetInitialPoints(x0)
  solver.SetStrictRanges(min=lb,max=ub)
  solver.SetEvaluationLimits(maxiter,maxfun)
  solver.SetEvaluationMonitor(evalmon)
  solver.SetGenerationMonitor(stepmon)

  tol = convergence_tol
  solver.Solve(cost, termination=NCOG(tol))

  solved_params = solver.bestSolution
  solved_energy = solver.bestEnergy
  func_evals = solver.evaluations
  return solved_params, solved_energy, func_evals


#######################################################################
# make a pseudo-global optimizer from a steepest descent optimizer
#######################################################################
def optimize(cost,lower,upper):
  from mystic.tools import random_seed
  from pyina.launchers import MoabSlurm as Pool
  random_seed(123)

  # generate arrays of points defining a grid in parameter space
  grid_dimensions = len(lower)
  bins = []
  for i in range(grid_dimensions):
    step = abs(upper[i] - lower[i])/nbins[i]
    bins.append( [lower[i] + (j+0.5)*step for j in range(nbins[i])] )

  # build a grid of starting points
  initial_values = gridpts(bins)

  # run optimizer for each grid point
  lb = [lower for i in range(len(initial_values))]
  ub = [upper for i in range(len(initial_values))]
  cf = [cost for i in range(len(initial_values))]
  # map:: params, energy, func_evals = local_optimize(cost,x0,lb,ub)
  config = {'queue':queue, 'timelimit':timelimit}
  results = Pool(nnodes, **config).map(local_optimize,cf,initial_values,lb,ub)
  #print("results = %s" % results)

  # get the results with the lowest energy
  best = list(results[0][0]), results[0][1]
  func_evals = results[0][2]
  for result in results[1:]:
    func_evals += result[2] # add function evaluations
    if result[1] < best[1]: # compare energy
      best = list(result[0]), result[1]

  # return best
  print("solved: %s" % best[0])
  scale = 1.0
  diameter_squared = -best[1] / scale  #XXX: scale != 0
  return diameter_squared, func_evals


#######################################################################
# loop over model parameters to calculate concentration of measure
#######################################################################
def UQ(start,end,lower,upper):
  diameters = []
  function_evaluations = []
  total_func_evals = 0
  total_diameter = 0.0

  for i in range(start,end+1):
    lb = lower + [lower[i]]
    ub = upper + [upper[i]]
    nbins[-1] = nbins[i] #XXX: kind of hackish
  
    #construct cost function and run optimizer
    cost = costFactory(i)
    subdiameter, func_evals = optimize(cost,lb,ub) #XXX: no init_condition

    function_evaluations.append(func_evals)
    diameters.append(subdiameter)

    total_func_evals += function_evaluations[-1]
    total_diameter += diameters[-1]

  print("subdiameters (squared): %s" % diameters)
  print("diameter (squared): %s" % total_diameter)
  print("func_evals: %s => %s" % (function_evaluations, total_func_evals))

  return total_diameter


#######################################################################
# rank, bounds, and restart information 
#######################################################################
if __name__ == '__main__':
  from math import sqrt

  function_name = "marc_surr"
  lower_bounds = [60.0, 0.0, 2.1]
  upper_bounds = [105.0, 30.0, 2.8]
# h = thickness = [60,105]
# a = obliquity = [0,30]
# v = speed = [2.1,2.8]

  RVstart = 0; RVend = 2
  RVmax = len(lower_bounds) - 1

  # when not a random variable, set the value to the lower bound
  for i in range(0,RVstart):
    upper_bounds[i] = lower_bounds[i]
  for i in range(RVend+1,RVmax+1):
    upper_bounds[i] = lower_bounds[i]

  lbounds = lower_bounds[RVstart:1+RVend]
  ubounds = upper_bounds[RVstart:1+RVend]

  #FIXME: these are *copies*... actual values contained in 'local_optimize'
  maxiter = 1000
  maxfun = 1e+6
  convergence_tol = 1e-4

  print("...SETTINGS...")
  print("nbins = %s" % nbins)
  print("maxiter = %s" % maxiter)
  print("maxfun = %s" % maxfun)
  print("convergence_tol = %s" % convergence_tol)
  #print("crossover = %s" % crossover)
  #print("percent_change = %s" % percent_change)
  print("..............\n\n")

  print(" model: f(x) = %s(x)" % function_name)
  param_string = "["
  for i in range(RVmax+1): 
    param_string += "'x%s'" % str(i+1)
    if i == (RVmax):
      param_string += "]"
    else:
      param_string += ", "

  print(" parameters: %s" % param_string)
  print("  varying 'xi', with i = %s" % list(range(RVstart+1,RVend+2)))
  print(" lower bounds: %s" % lower_bounds)
  print(" upper bounds: %s" % upper_bounds)
# print(" ...")
  nbins.append(None) #XXX: kind of hackish
  diameter = UQ(RVstart,RVend,lower_bounds,upper_bounds)

# EOF



#!/usr/bin/env python
#
# Author: Alta Fang (altafang @caltech and alta @princeton)
# Author: Mike McKerns (mmckerns @caltech and @uqfoundation)
# Copyright (c) 1997-2016 California Institute of Technology.
# Copyright (c) 2016-2022 The Uncertainty Quantification Foundation.
# License: 3-clause BSD.  The full license text is available at:
#  - https://github.com/uqfoundation/mystic/blob/master/LICENSE
"""Test Mystic's performance on some benchmark problems, with Mystic's settings
adjusted to achieve the best results.
"""
from mystic.monitors import Monitor
from mystic.math import almostEqual
from mystic.tools import random_seed
random_seed(123)
import time

verbose = False


def test_rosenbrock(verbose=False):
    """Test the 2-dimensional Rosenbrock function.

Testing 2-D Rosenbrock:
Expected: x=[1., 1.] and f=0

Using DifferentialEvolutionSolver:
Solution:  [ 1.00000037  1.0000007 ]
f value:  2.29478683682e-13
Iterations:  99
Function evaluations:  3996
Time elapsed:  0.582273006439  seconds

Using DifferentialEvolutionSolver2:
Solution:  [ 0.99999999  0.99999999]
f value:  3.84824937598e-15
Iterations:  100
Function evaluations:  4040
Time elapsed:  0.577210903168  seconds

Using NelderMeadSimplexSolver:
Solution:  [ 0.99999921  1.00000171]
f value:  1.08732211477e-09
Iterations:  70
Function evaluations:  130
Time elapsed:  0.0190329551697  seconds

Using PowellDirectionalSolver:
Solution:  [ 1.  1.]
f value:  0.0
Iterations:  28
Function evaluations:  859
Time elapsed:  0.113857030869  seconds
"""

    if verbose:
        print("Testing 2-D Rosenbrock:")
        print("Expected: x=[1., 1.] and f=0")
    from mystic.models import rosen as costfunc
    ndim = 2
    lb = [-5.]*ndim
    ub = [5.]*ndim
    x0 = [2., 3.]
    maxiter = 10000
    
    # DifferentialEvolutionSolver
    if verbose: print("\nUsing DifferentialEvolutionSolver:")
    npop = 40
    from mystic.solvers import DifferentialEvolutionSolver
    from mystic.termination import ChangeOverGeneration as COG
    from mystic.strategy import Rand1Bin
    esow = Monitor()
    ssow = Monitor() 
    solver = DifferentialEvolutionSolver(ndim, npop)
    solver.SetInitialPoints(x0)
    solver.SetStrictRanges(lb, ub)
    solver.SetEvaluationLimits(generations=maxiter)
    solver.SetEvaluationMonitor(esow)
    solver.SetGenerationMonitor(ssow)
    term = COG(1e-10)
    time1 = time.time() # Is this an ok way of timing?
    solver.Solve(costfunc, term, strategy=Rand1Bin)
    sol = solver.Solution()
    time_elapsed = time.time() - time1
    fx = solver.bestEnergy
    if verbose:
        print("Solution: %s" % sol)
        print("f value: %s" % fx)
        print("Iterations: %s" % solver.generations)
        print("Function evaluations: %s" % len(esow.x))
        print("Time elapsed: %s seconds" % time_elapsed)
    assert almostEqual(fx, 2.29478683682e-13, tol=3e-3)

    # DifferentialEvolutionSolver2
    if verbose: print("\nUsing DifferentialEvolutionSolver2:")
    npop = 40
    from mystic.solvers import DifferentialEvolutionSolver2
    from mystic.termination import ChangeOverGeneration as COG
    from mystic.strategy import Rand1Bin
    esow = Monitor()
    ssow = Monitor() 
    solver = DifferentialEvolutionSolver2(ndim, npop)
    solver.SetInitialPoints(x0)
    solver.SetStrictRanges(lb, ub)
    solver.SetEvaluationLimits(generations=maxiter)
    solver.SetEvaluationMonitor(esow)
    solver.SetGenerationMonitor(ssow)
    term = COG(1e-10)
    time1 = time.time() # Is this an ok way of timing?
    solver.Solve(costfunc, term, strategy=Rand1Bin)
    sol = solver.Solution()
    time_elapsed = time.time() - time1
    fx = solver.bestEnergy
    if verbose:
        print("Solution: %s" % sol)
        print("f value: %s" % fx)
        print("Iterations: %s" % solver.generations)
        print("Function evaluations: %s" % len(esow.x))
        print("Time elapsed: %s seconds" % time_elapsed)
    assert almostEqual(fx, 3.84824937598e-15, tol=3e-3)

    # NelderMeadSimplexSolver
    if verbose: print("\nUsing NelderMeadSimplexSolver:")
    from mystic.solvers import NelderMeadSimplexSolver
    from mystic.termination import CandidateRelativeTolerance as CRT
    esow = Monitor()
    ssow = Monitor() 
    solver = NelderMeadSimplexSolver(ndim)
    solver.SetInitialPoints(x0)
    solver.SetStrictRanges(lb, ub)
    solver.SetEvaluationLimits(generations=maxiter)
    solver.SetEvaluationMonitor(esow)
    solver.SetGenerationMonitor(ssow)
    term = CRT()
    time1 = time.time() # Is this an ok way of timing?
    solver.Solve(costfunc, term)
    sol = solver.Solution()
    time_elapsed = time.time() - time1
    fx = solver.bestEnergy
    if verbose:
        print("Solution: %s" % sol)
        print("f value: %s" % fx)
        print("Iterations: %s" % solver.generations)
        print("Function evaluations: %s" % len(esow.x))
        print("Time elapsed: %s seconds" % time_elapsed)
    assert almostEqual(fx, 1.08732211477e-09, tol=3e-3)

    # PowellDirectionalSolver
    if verbose: print("\nUsing PowellDirectionalSolver:")
    from mystic.solvers import PowellDirectionalSolver
    from mystic.termination import NormalizedChangeOverGeneration as NCOG
    esow = Monitor()
    ssow = Monitor() 
    solver = PowellDirectionalSolver(ndim)
    solver.SetInitialPoints(x0)
    solver.SetStrictRanges(lb, ub)
    solver.SetEvaluationLimits(generations=maxiter)
    solver.SetEvaluationMonitor(esow)
    solver.SetGenerationMonitor(ssow)
    term = NCOG(1e-10)
    time1 = time.time() # Is this an ok way of timing?
    solver.Solve(costfunc, term)
    sol = solver.Solution()
    time_elapsed = time.time() - time1
    fx = solver.bestEnergy
    if verbose:
        print("Solution: %s" % sol)
        print("f value: %s" % fx)
        print("Iterations: %s" % solver.generations)
        print("Function evaluations: %s" % len(esow.x))
        print("Time elapsed: %s seconds" % time_elapsed)
    assert almostEqual(fx, 0.0, tol=3e-3)

def test_griewangk(verbose=False):
    """Test Griewangk's function, which has many local minima.

Testing Griewangk:
Expected: x=[0.]*10 and f=0

Using DifferentialEvolutionSolver:
Solution:  [  8.87516194e-09   7.26058147e-09   1.02076001e-08   1.54219038e-08
  -1.54328461e-08   2.34589663e-08   2.02809360e-08  -1.36385836e-08
   1.38670373e-08   1.59668900e-08]
f value:  0.0
Iterations:  4120
Function evaluations:  205669
Time elapsed:  34.4936850071  seconds

Using DifferentialEvolutionSolver2:
Solution:  [ -2.02709316e-09   3.22017968e-09   1.55275472e-08   5.26739541e-09
  -2.18490470e-08   3.73725584e-09  -1.02315312e-09   1.24680355e-08
  -9.47898116e-09   2.22243557e-08]
f value:  0.0
Iterations:  4011
Function evaluations:  200215
Time elapsed:  32.8412370682  seconds
"""

    if verbose:
        print("Testing Griewangk:")
        print("Expected: x=[0.]*10 and f=0")
    from mystic.models import griewangk as costfunc
    ndim = 10
    lb = [-400.]*ndim
    ub = [400.]*ndim
    maxiter = 10000
    seed = 123 # Re-seed for each solver to have them all start at same x0
    
    # DifferentialEvolutionSolver
    if verbose: print("\nUsing DifferentialEvolutionSolver:")
    npop = 50
    random_seed(seed)
    from mystic.solvers import DifferentialEvolutionSolver
    from mystic.termination import ChangeOverGeneration as COG
    from mystic.termination import CandidateRelativeTolerance as CRT
    from mystic.termination import VTR
    from mystic.strategy import Rand1Bin, Best1Bin, Rand1Exp
    esow = Monitor()
    ssow = Monitor() 
    solver = DifferentialEvolutionSolver(ndim, npop)
    solver.SetRandomInitialPoints(lb, ub)
    solver.SetStrictRanges(lb, ub)
    solver.SetEvaluationLimits(generations=maxiter)
    solver.SetEvaluationMonitor(esow)
    solver.SetGenerationMonitor(ssow)
    solver.enable_signal_handler()
    #term = COG(1e-10)
    #term = CRT()
    term = VTR(0.)
    time1 = time.time() # Is this an ok way of timing?
    solver.Solve(costfunc, term, strategy=Rand1Exp, \
                 CrossProbability=0.3, ScalingFactor=1.0)
    sol = solver.Solution()
    time_elapsed = time.time() - time1
    fx = solver.bestEnergy
    if verbose:
        print("Solution: %s" % sol)
        print("f value: %s" % fx)
        print("Iterations: %s" % solver.generations)
        print("Function evaluations: %s" % len(esow.x))
        print("Time elapsed: %s seconds" % time_elapsed)
    assert almostEqual(fx, 0.0, tol=3e-3)

    # DifferentialEvolutionSolver2
    if verbose: print("\nUsing DifferentialEvolutionSolver2:")
    npop = 50
    random_seed(seed)
    from mystic.solvers import DifferentialEvolutionSolver2
    from mystic.termination import ChangeOverGeneration as COG
    from mystic.termination import CandidateRelativeTolerance as CRT
    from mystic.termination import VTR
    from mystic.strategy import Rand1Bin, Best1Bin, Rand1Exp
    esow = Monitor()
    ssow = Monitor() 
    solver = DifferentialEvolutionSolver2(ndim, npop)
    solver.SetRandomInitialPoints(lb, ub)
    solver.SetStrictRanges(lb, ub)
    solver.SetEvaluationLimits(generations=maxiter)
    solver.SetEvaluationMonitor(esow)
    solver.SetGenerationMonitor(ssow)
    #term = COG(1e-10)
    #term = CRT()
    term = VTR(0.)
    time1 = time.time() # Is this an ok way of timing?
    solver.Solve(costfunc, term, strategy=Rand1Exp, \
                 CrossProbability=0.3, ScalingFactor=1.0)
    sol = solver.Solution()
    time_elapsed = time.time() - time1
    fx = solver.bestEnergy
    if verbose:
        print("Solution: %s" % sol)
        print("f value: %s" % fx)
        print("Iterations: %s" % solver.generations)
        print("Function evaluations: %s" % len(esow.x))
        print("Time elapsed: %s seconds" % time_elapsed)
    assert almostEqual(fx, 0.0, tol=3e-3)

if __name__ == '__main__':
    test_rosenbrock(verbose)
   #test_griewangk(verbose)



#!/usr/bin/env python
#
# Author: Alta Fang (altafang @caltech and alta @princeton)
# Author: Mike McKerns (mmckerns @caltech and @uqfoundation)
# Copyright (c) 1997-2016 California Institute of Technology.
# Copyright (c) 2016-2022 The Uncertainty Quantification Foundation.
# License: 3-clause BSD.  The full license text is available at:
#  - https://github.com/uqfoundation/mystic/blob/master/LICENSE
"""
Testing the 'Peaks" Function.

(tests VTR when minimum has negative value)
"""
from math import *
from mystic.models import peaks

nd = 2
npop = 20
tol = 0.05
lb = [-3.]*nd
ub = [3.]*nd

from mystic.tools import random_seed
#random_seed(123)
from mystic.differential_evolution import DifferentialEvolutionSolver
from mystic.termination import VTR
from mystic.termination import ChangeOverGeneration as COG

solver = DifferentialEvolutionSolver(nd, npop)
solver.SetRandomInitialPoints(lb, ub)
solver.SetStrictRanges(lb, ub)
term = VTR(tol) 
#term = COG()
solver.Solve(peaks, term, disp=True)
sol = solver.Solution()
print('solution = %s' % sol)
print('expected = [0.23, -1.63]')

try:
    from scipy.stats import uniform
except ImportError:
    exit()
from mystic.math import Distribution

print('-'*60)
print('using a uniform distribution...')
solver = DifferentialEvolutionSolver(nd, npop)
solver.SetSampledInitialPoints(Distribution(uniform, lb[0], ub[0]))
solver.SetStrictRanges(lb, ub)
term = VTR(tol) 
#term = COG()
solver.Solve(peaks, term, disp=True)
sol = solver.Solution()
print('solution = %s' % sol)
print('expected = [0.23, -1.63]')






#!/usr/bin/env python
#
# Author: Mike McKerns (mmckerns @caltech and @uqfoundation)
# Copyright (c) 1997-2016 California Institute of Technology.
# Copyright (c) 2016-2022 The Uncertainty Quantification Foundation.
# License: 3-clause BSD.  The full license text is available at:
#  - https://github.com/uqfoundation/mystic/blob/master/LICENSE
"""
tools related to sampling
"""
# These functions are consolidated and adapted from samples.py and cut.py in 
# branches/UQ/math/ for general use.
# sample() is from cut.py, and samplepts() is from examples2/seesaw2d.py, and 
# everything else is from samples.py

# SAMPLING #
def _random_samples(lb,ub,npts=10000):
  """
generate npts random samples between given lb & ub

Inputs:
    lower bounds  --  a list of the lower bounds
    upper bounds  --  a list of the upper bounds
    npts  --  number of sample points [default = 10000]
"""
  from mystic.tools import random_state
  dim = len(lb)
  pts = random_state(module='numpy.random').rand(dim,npts)
  for i in range(dim):
    pts[i] = (pts[i] * abs(ub[i] - lb[i])) + lb[i]
  return pts  #XXX: returns a numpy.array
 #return [list(i) for i in pts]


def random_samples(lb,ub, npts=10000, dist=None, clip=False):
  """
generate npts samples from the given distribution between given lb & ub

Inputs:
    dist  --  a mystic.tools.Distribution instance (or list of Distributions)
    lower bounds  --  a list of the lower bounds
    upper bounds  --  a list of the upper bounds
    npts  --  number of sample points [default = 10000]
    clip  --  if True, clip at bounds, else resample [default = False]
"""
  if dist is None:
    return _random_samples(lb,ub, npts)
  import numpy as np
  if hasattr(dist, '__len__'): #FIXME: isiterable
    pts = np.array(tuple(di(npts) for di in dist)).T
  else:
    pts = dist((npts,len(lb))) # transpose of desired shape
    dist = (dist,)*len(lb)
  pts = np.clip(pts, lb, ub).T
  if clip: return pts  #XXX: returns a numpy.array
  bad = ((pts.T == lb) + (pts.T == ub)).T
  new = bad.sum(-1)
  _n, n = 1, 1000 #FIXME: fixed number of max tries
  while any(new):
    if _n == n: #XXX: slows the while loop...
      raise RuntimeError('bounds could not be applied in %s iterations' % n)
    for i,inew in enumerate(new): #XXX: slows... but enables iterable dist
      if inew: pts[i][bad[i]] = dist[i](inew)
    pts = np.clip(pts.T, lb, ub).T
    bad = ((pts.T == lb) + (pts.T == ub)).T
    new = bad.sum(-1)
    _n += 1
  return pts  #XXX: returns a numpy.array


def sample(f,lb,ub,npts=10000):
  """
return number of failures and successes for some boolean function f

Inputs:
    f -- a function that returns True for 'success' and False for 'failure'
    lb -- a list of lower bounds
    ub -- a list of upper bounds
    npts -- the number of points to sample [Default is npts=10000]
"""
  from numpy import transpose
  pts = _random_samples(lb, ub, npts)

  failure = 0; success = 0
  for i in range(npts):
    xvector = transpose(pts)[i]
    if f(list(xvector)):
      success += 1
    else:
      failure += 1
  return failure,success


# STATISTICS #
def sampled_mean(f, lb,ub, npts=10000):
  """
use random sampling to calculate the mean of a function

Inputs:
    f -- a function that takes a list and returns a number
    lb -- a list of lower bounds
    ub -- a list of upper bounds
    npts -- the number of points to sample [Default is npts=10000]
"""
  from numpy import inf, transpose
  from mystic.tools import wrap_bounds
  pts = _random_samples(lb, ub, npts)
  f = wrap_bounds(f,lb,ub)
  ave = 0; count = 0
  for i in range(len(pts[0])):
    if len(lb) != 1:
      xvector = transpose(pts)[i]
      Fx = f(list(xvector))
    else:
      Fx = f(float(transpose(pts)[i]))
    if Fx != -inf: # outside of bounds evaluates to -inf
      ave += Fx
      count += 1
  if not count: return None  #XXX: define 0/0 = None
  ave = float(ave) / float(count)
  return ave


def sampled_variance(f, lb, ub, npts=10000): #XXX: this could be improved
  """
use random sampling to calculate the variance of a function

Inputs:
    f -- a function that takes a list and returns a number
    lb -- a list of lower bounds
    ub -- a list of upper bounds
    npts -- the number of points to sample [Default is npts=10000]
"""
  m = sampled_mean(f,lb,ub)
  def g(x):
    return abs(f(x) - m)**2
  return sampled_mean(g,lb,ub)


def sampled_pof(f, lb, ub, npts=10000):
  """
use random sampling to calculate probability of failure for a function

Inputs:
    f -- a function that returns True for 'success' and False for 'failure'
    lb -- a list of lower bounds
    ub -- a list of upper bounds
    npts -- the number of points to sample [Default is npts=10000]
"""
  pts = _random_samples(lb, ub, npts)
  return _pof_given_samples(f, pts)


# ALTERNATE: GIVEN SAMPLE POINTS #
def _pof_given_samples(f, pts):
  """
use given sample pts to calculate probability of failure for function f

Inputs:
    f -- a function that returns True for 'success' and False for 'failure'
    pts -- a list of sample points
"""
  from numpy import transpose
  failure = 0
  npts = len(pts[0]) #XXX: fails when pts = []; also assumes a nested list
  for i in range(npts):
    xvector = transpose(pts)[i]
    if not f(list(xvector)):
      failure += 1
  pof = float(failure) / float(npts)
  return pof


def _minimum_given_samples(f, pts):
  """
use given sample pts to calculate minimum for function f

Inputs:
    f -- a function that returns a single value, given a list of inputs
    pts -- a list of sample points
"""
  from numpy import transpose
  return min([f(list(i)) for i in transpose(pts)])


def _expectation_given_samples(f, pts):
  """
use given sample pts to calculate expected value for function f

Inputs:
    f -- a function that returns a single value, given a list of inputs
    pts -- a list of sample points
"""
  from numpy import transpose, mean
  return mean([f(list(i)) for i in transpose(pts)])


def _maximum_given_samples(f, pts):
  """
use given sample pts to calculate maximum for function f

Inputs:
    f -- a function that returns a single value, given a list of inputs
    pts -- a list of sample points
"""
  from numpy import transpose
  return max([f(list(i)) for i in transpose(pts)])


def _ptp_given_samples(f, pts):
  """
use given sample pts to calculate spread for function f

Inputs:
    f -- a function that returns a single value, given a list of inputs
    pts -- a list of sample points
"""
  from numpy import transpose, ptp
  return ptp([f(list(i)) for i in transpose(pts)])


def sampled_pts(pts,lb,ub):
  """
determine the number of sample points inside the given bounds

Inputs:
    pts -- a list of sample points
    lb -- a list of lower bounds
    ub -- a list of upper bounds
"""
  from numpy import inf, transpose
  def identity(x):
    return x
  from mystic.tools import wrap_bounds
  f = wrap_bounds(identity,lb,ub)
  npts = 0
  for i in range(len(pts[0])):
    xvector = transpose(pts)[i]
    Fx = f(list(xvector))
    if Fx != -inf: # outside of bounds evaluates to -inf
      npts += 1
  return npts

def sampled_prob(pts,lb,ub):
  """
calculates probability by sampling if points are inside the given bounds

Inputs:
    pts -- a list of sample points
    lb -- a list of lower bounds
    ub -- a list of upper bounds
"""
  prob = float(sampled_pts(pts,lb,ub)) / float(len(pts[0]))
  return prob






#def minF(x):
#  return model(x)

#def maxF(x):
#  return -model(x)

def alpha(n,diameter,epsilon=0.01):
  from math import log
 #return diameter * n**(-0.5) * (-log(epsilon))**(0.5)
  return diameter * (-log(epsilon) / (2.0 * n))**(0.5)


#######################################################################


if __name__ == '__main__':

  def __test1():
    # From branches/UQ/math/samples.py
    num_sample_points = 5
    lower = [10.0, 0.0, 2.1]
    upper = [100.0, 30.0, 2.8]

    pts = _random_samples(lower,upper,num_sample_points)
    print("randomly sampled points\nbetween %s and %s" % (lower, upper))
    print(pts)

  def __test2():
    # From branches/UQ/math/cut.py
    from mystic.tools import random_seed
    random_seed(123)
    lower = [-60.0, -10.0, -50.0]
    upper = [105.0, 30.0, 75.0]

    def model(x):
      x1,x2,x3 = x
      if x1 > (x2 + x3): return x1*x2 - x3
      return 0.0

    failure,success = sample(model,lower,upper)
    pof = float(failure) / float(failure + success)
    print("PoF using method 1: %s" % pof)
    random_seed(123)
    print("PoF using method 2: %s" % sampled_pof(model,lower,upper))

    # run the tests
    __test1()
    __test2()


# EOF
